#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_charset.h"
#include "udm_spell.h"
#include "udm_proto.h"
#include "udm_cache.h"
#include "udm_crc32.h"
#include "udm_indexer.h"
#include "udm_db.h"
#include "udm_boolean.h"
#include "udm_searchtool.h"

/*
#define DEBUG_SEARCH 1
*/

#define LOGDIR	"raw"
#define TREEDIR	"tree"

/**********************************/

typedef struct udm_log_struct {
	int	stamp;
	int	category;
	int	url_id;
	int	wrd_id;
} UDM_LOGWORD;

typedef struct udm_cache_table {
	int wrd_id;
	int pos;
	int len;
} UDM_CACHETABLE;

typedef struct udm_cache_hheader {
	int ntables;
	int version;
} UDM_CACHEHEADER;

typedef struct udm_cacheword_struct {
	int	url_id;
	int	category;
} UDM_CACHEWORD;


/******** Convert category string into 32 bit number *************/
static void UdmDecodeCatStr(char * cat_str, int * cat, int * mask){
unsigned int t[5];
char str[128]="";

	strcpy(str,cat_str);
	strcat(str,"000000000000");
	str[10]=0;
	sscanf(str,"%02x%02x%02x%02x%02x",t+0,t+1,t+2,t+3,t+4);

	*mask= (int)( (t[0]?0x7F<<25:0) | (t[1]?0x7F<<18:0) | 
		(t[2]?0x3F<<12:0) | (t[3]?0x3F<<6:0) | (t[4]?0x3F<<0:0));

	*cat= (int)((t[0]<<25) | (t[1]<<18) | (t[2]<<12) | (t[3]<<6) | (t[4]<<0));
}

/*************************** Sort functions **************************/
/* Sort SEARCHWORD by url_id order */
static int cmpurlid(const void *s1,const void *s2){
	return(((UDM_SEARCHWORD*)s1)->url_id-((UDM_SEARCHWORD*)s2)->url_id);
}

/* Function to sort CRCWORD list in URL_ID order */
static int cmpcrc(const void *s1,const void *s2){
int n1,n2;
	n1=(unsigned char)((unsigned int)(((UDM_CRCWORD*)s1)->wrd_id)>>24);
	n2=(unsigned char)((unsigned int)(((UDM_CRCWORD*)s2)->wrd_id)>>24);
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Function to sort LOGWORD list in (WRD_ID,TIME_STAMP) order */
static int cmplog(const void *s1,const void *s2){
unsigned int n1,n2;
	n1=((UDM_LOGWORD*)s1)->wrd_id;
	n2=((UDM_LOGWORD*)s2)->wrd_id;
	if(n1==n2){
		n1=((UDM_LOGWORD*)s1)->url_id;
		n2=((UDM_LOGWORD*)s2)->url_id;
		if(n1==n2){
			n1=((UDM_LOGWORD*)s2)->stamp;
			n2=((UDM_LOGWORD*)s1)->stamp;
		}
	}
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Function to sort LOGWORD list in URL_ID order */
static int cmpurllog(const void *s1,const void *s2){
unsigned int n1,n2;
	n1=((UDM_LOGWORD*)s1)->url_id;
	n2=((UDM_LOGWORD*)s2)->url_id;
	if(n1==n2){
		n1=((UDM_LOGWORD*)s2)->stamp;
		n2=((UDM_LOGWORD*)s1)->stamp;
	}
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Function to sort CACHEWORD list in URL_ID order */
static int cmpcache(const void *s1,const void *s2){
int n1,n2;
	n1=((UDM_LOGWORD*)s1)->wrd_id;
	n2=((UDM_LOGWORD*)s2)->wrd_id;
	if(n1==n2){
		n1=((UDM_LOGWORD*)s1)->url_id;
		n2=((UDM_LOGWORD*)s2)->url_id;
	}
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Sort tables by their length order */
static int cmptable(const void * t1, const void * t2){
	return(((UDM_CACHETABLE*)t2)->len-((UDM_CACHETABLE*)t1)->len);
}

/******* These two functions open/close all 256+1 logs ***********/

int UdmOpenCache(UDM_AGENT * Indexer,char * errmsg){
int i;
char fname[UDMSTRSIZ];
	if(Indexer->crcdict[0]>=0){
		return(IND_OK);
	}
	for(i=0;i<257;i++){
		if(i<256){
			sprintf(fname,"%s%c%s%c%02X.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH,i);
		}else{
			sprintf(fname,"%s%c%s%cdel.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH);
		}
		if((Indexer->crcdict[i]=open(fname,O_RDWR|O_CREAT|O_APPEND,0644))<0){
			sprintf(errmsg,"Can't open \"%s\" for writing: %s",fname,strerror(errno));
			UdmCloseCache(Indexer);
			return(IND_ERROR);
		}
	}
	return(IND_OK);
}

void UdmCloseCache(UDM_AGENT * Indexer){
int i;
	for(i=0;i<257;i++){
		if(Indexer->crcdict[i]>0)
			close(Indexer->crcdict[i]);
	}
}

/**************** This function is currently not used ***************/
int PresentInLog(UDM_LOGWORD *buf, int buf_count, const int url_id){
register int m;
register int l;
register int r;

	l=0;
	r=buf_count;
	while(l<r){
		m=(l+r)/2;
		if((unsigned int)buf[m].url_id<(unsigned int)url_id) l=m+1;
		else r=m;
	}
	if(r==buf_count) return(0);
	else
		if(buf[r].url_id==url_id) return(buf[r].stamp);
		else return(0);
}

int PresentInInt(int *buf, int buf_count, const int url_id){
register int m;
register int l;
register int r;

	l=0;
	r=buf_count;
	while(l<r){
		m=(l+r)/2;
		if((unsigned int)buf[m]<(unsigned int)url_id) l=m+1;
		else r=m;
	}
	if(r==buf_count) return(0);
	else
		if(buf[r]==url_id) return(url_id);
		else return(0);
}


int UdmDeleteFromCache(char *fname, UDM_LOGWORD *del_buf, int del_count){
	int oldfd,newfd;
	UDM_CACHEHEADER	header;
	UDM_CACHEHEADER	newheader;
	UDM_CACHETABLE	table[4096];
	UDM_CACHETABLE	newtable[4096];
	UDM_CACHEWORD	*cache;
	int i,data_len;
	char tmpname[UDMSTRSIZ]="";

	if((oldfd=open(fname,O_RDONLY))>=0){
		read(oldfd,&header,sizeof(header));
		read(oldfd,table,header.ntables*sizeof(UDM_CACHETABLE));
		data_len=0;
		for(i=0;i<header.ntables;i++)data_len+=table[i].len;
		cache=(UDM_CACHEWORD*)malloc(data_len);
		read(oldfd,cache,data_len);
		close(oldfd);

		printf("Delete from cache-file %s\n",fname);

		for(i=0;i<header.ntables;i++){
			int pos,len,j;
			int last,from;

			pos=table[i].pos/sizeof(UDM_CACHEWORD);
			len=table[i].len/sizeof(UDM_CACHEWORD);

			/* Make holes */
			for(j=pos;j<pos+len;j++){
				if(PresentInLog(del_buf,del_count,cache[j].url_id)){
					cache[j].url_id=0;
					table[i].len-=sizeof(UDM_CACHEWORD);
				}
				/*
				for(k=0;k<del_count;k++){
					
					if(cache[j].url_id==del_buf[k].url_id){
						cache[j].url_id=0;
						table[i].len-=sizeof(UDM_CACHEWORD);
						break;
					}
				}
				*/				
			}

			/* Turn off holes */
			if(table[i].len>0){
				j=pos;
				while((cache[j].url_id!=0)&&(j<pos+len))j++;
				last=j;
 				while(j<pos+len){
					while((cache[j].url_id==0)&&(j<pos+len))j++;
					from=j;
					while((cache[j].url_id!=0)&&(j<pos+len))j++;
					memcpy(&cache[last],&cache[from],(j-from)*sizeof(UDM_CACHEWORD));
					last+=j-from; 
				}
			}
		}
		

		/* Create new table */
		data_len=0;
		newheader.ntables=0;
		for(i=0;i<header.ntables;i++){
			if(table[i].len>0){
				newtable[newheader.ntables].wrd_id=table[i].wrd_id;
				newtable[newheader.ntables].pos=data_len;
				newtable[newheader.ntables].len=table[i].len;
				data_len+=newtable[newheader.ntables].len;
				newheader.ntables++;
			}
		}

		if(newheader.ntables==0){
			unlink(fname);
		}else{
			/* Write new cache file */	
			sprintf(tmpname,"%s.tmp",fname);
			newfd=open(tmpname,O_WRONLY|O_CREAT|O_TRUNC,0644);
			write(newfd,&newheader,sizeof(newheader));
			write(newfd,newtable,newheader.ntables*sizeof(UDM_CACHETABLE));
			write(newfd,cache,data_len);
			close(newfd);
			rename(tmpname,fname);
		}

		UDM_FREE(cache);

		return(0);
	}else{
		return(1);
	}
}


/***** Write a marker that the old content of url_id should be deleted ****/
int UdmDeleteURLFromCache(UDM_AGENT * Indexer,int url_id){
	UDM_LOGWORD logwrd;
	int i;

	/* Check whether XX.log files are already open */
	if(Indexer->crcdict[0]<0){
		/* Not open yet, let's open it now */
		i=UdmOpenCache(Indexer,UdmAgentErrorMsg(Indexer));
		if(i!=IND_OK){
			UdmCloseCache(Indexer);
			return(i);
		}
	}
	logwrd.stamp=time(NULL);
	logwrd.category=0;
	logwrd.url_id=url_id;
	logwrd.wrd_id=0;

	if((sizeof(logwrd)!=write(Indexer->crcdict[256],&logwrd,sizeof(logwrd)))){
		sprintf(UdmAgentErrorMsg(Indexer),"Can't write to delete log: %s",strerror(errno));
		UdmCloseCache(Indexer);
		return(IND_ERROR);
	}
	return(IND_OK);
}


/******** Write words into 256+1 logs **************************/
int UdmStoreWordsCache(UDM_AGENT * Indexer,int url_id,char * catstr, int status){
int fd,i,bytes,stamp;
int category,mask;
UDM_CRCWORD * cw;

	/* Checke whether XX.log files are already open */
	if(Indexer->crcdict[0]<0){
		/* Not open yet, let's open it now */
		i=UdmOpenCache(Indexer,UdmAgentErrorMsg(Indexer));
		if(i!=IND_OK){
			UdmCloseCache(Indexer);
			return(i);
		}
	}
	
	/* Convert category string into 32 bit number */
	UdmDecodeCatStr(catstr,&category,&mask);

	/* Mark that old content should be deleted */
	if(status){
		i=UdmDeleteURLFromCache(Indexer,url_id);
		if(i!=IND_OK)return(i);
	}

	stamp=time(NULL);

	if(Indexer->nwords){
		int prev,first=0;
		bytes=sizeof(UDM_CRCWORD)*Indexer->nwords;
		cw=(UDM_CRCWORD*)malloc(bytes);
		
		for(i=0;i<Indexer->nwords;i++){
			cw[i].url_id=url_id;
			cw[i].weight=Indexer->Word[i].count,
			cw[i].wrd_id=UdmStrCRC32(Indexer->Word[i].word);
			//printf("%08X\t%s\n",cw[i].wrd_id,Indexer->Word[i].word);
		}
		qsort((void*)cw,Indexer->nwords,sizeof(UDM_CRCWORD),cmpcrc);
		
		prev=((unsigned int)cw[0].wrd_id)>>24;
		for(i=1;i<Indexer->nwords+1;i++){
			int n;
			
			if(i<Indexer->nwords+1)n=((unsigned int)cw[i].wrd_id)>>24;
			else n=prev+1;
			
			if((n!=prev)||(i==Indexer->nwords)){
			
				if((first!=i)){
					int w,wcount;
					UDM_LOGWORD * buf;
					
					fd=Indexer->crcdict[prev];
					wcount=i-first;
					
					bytes=wcount*sizeof(UDM_LOGWORD);
					buf=(UDM_LOGWORD *)malloc(bytes);
					for(w=0;w<wcount;w++){
						//printf("i: %d first: %d %s\n",i,first,Indexer->Word[first+w].word);
						buf[w].stamp=stamp;
						buf[w].category=category;
						buf[w].url_id=cw[first+w].url_id;
						buf[w].wrd_id=(cw[first+w].wrd_id<<8)+(cw[first+w].weight&0xFF);
						//printf("%08X %08X\n",buf[w].url_id,buf[w].wrd_id);
					}
					//printf("Write: %2X %d bytes\n",prev,bytes);
					if((bytes!=write(fd,(void *)buf,bytes))){
						free(buf);
						sprintf(UdmAgentErrorMsg(Indexer),"Can't write to log file #%d: %s",prev,strerror(errno));
						UdmCloseCache(Indexer);
						return(IND_ERROR);
					}
					free(buf);
				}
				first=i;
			}
			prev=n;
		}
		free(cw);
	}	
	return(IND_OK);
}

/***************** Split one log into cache *****************/

/* 
This function removes duplicate
(wrd_id,url_id) pairs using timestamps.
Most new pairs will remane.
*/

static int RemoveURLDups(UDM_LOGWORD *words, int n){
	int i,j;
	
	j=0;
	for(i=1;i<n;i++){
		if(words[j].url_id!=words[i].url_id){
			j++;
			if(i!=j)words[j]=words[i];
		}
	}
	return(j+1);
}

/* 
This function removes non-fresh  
words from UDM_LOGWORD array using timestamps
from delete log. Only those (word_id,url_id)
pairs which are newer than (url_id) from 
delete log will remane in "words" array 
*/

static int RemoveOldWords(UDM_LOGWORD *words, int n,UDM_LOGWORD *del, int del_count){
	int i,j;
	
	j=0;
	for(i=1;i<n;i++){
		if((words[j].wrd_id!=words[i].wrd_id)||(words[j].url_id!=words[i].url_id)){
			if(PresentInLog(del,del_count,words[i].url_id)<=words[i].stamp){
				j++;
				if(i!=j){
					words[j]=words[i];
				}
			}
		}
	}
	/* Check first word */
	if(j==0){
		if(PresentInLog(del,del_count,words[0].url_id)>words[0].stamp)
			return(0);
	}
	return(j+1);
}

int UdmClearCacheTree(){
int i,j,k;
char fname[UDMSTRSIZ];
	for(i=0;i<256;i++)
	for(j=0;j<16;j++)
	for(k=0;k<256;k++){
		sprintf(fname,"%s%c%s%c%02X%c%01X%c%02X%01X%02X000",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,i,UDMSLASH,j,UDMSLASH,i,j,k);
		printf("%s\n",fname);
		unlink(fname);
	}
	return(0);
}

int UdmSplitCacheLog(int log){
int fd,dd,bytes,i,bufsize;
char fname[UDMSTRSIZ]="";
char dname[UDMSTRSIZ]="";
UDM_LOGWORD * buf;
UDM_LOGWORD * del_buf;
int del_count;
int done[4096];
struct stat sb;

	/* Open log file */
	sprintf(fname,"%s%c%s%c%02X.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH,log);
	if((fd=open(fname,O_RDONLY))<0){
		return(IND_ERROR);
	}
	
	/* Open del log file */
	sprintf(dname,"%s%c%s%cdel.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH);
	if((dd=open(dname,O_RDONLY))<0){
		return(IND_ERROR);
	}

	/* Allocate del buffer */
	fstat(dd, &sb);
	del_buf=(UDM_LOGWORD*)malloc(sb.st_size);
	del_count=read(dd,del_buf,sb.st_size)/sizeof(UDM_LOGWORD);
	close(dd);

	/* Remove duplicates URLs in DEL log     */
	/* Keep only oldest records for each URL */
	qsort(del_buf,del_count,sizeof(UDM_LOGWORD),cmpurllog);
	del_count=RemoveURLDups(del_buf,del_count);

	/* Allocate buffer for log */
	bufsize=sizeof(UDM_LOGWORD)*1024*1024*1; 
	buf=(UDM_LOGWORD*)malloc(bufsize);
	
	/* Zero all Markers */
	UDMMEMZERO(done,sizeof(done));


	while((bytes=read(fd,buf,bufsize))){
		int n,new,prev,first=0;
		n=bytes/sizeof(UDM_LOGWORD);
		printf("Load %d words (%d/%d bytes)\n",n,bytes,bufsize);
		qsort(buf,n,sizeof(UDM_LOGWORD),cmplog);
		printf("N before: %d\n",n);
		n=RemoveOldWords(buf,n,del_buf,del_count);
		printf("N after : %d\n",n);

		prev=((unsigned int)buf[0].wrd_id)>>8;

		for(i=1;i<n+1;i++){
			if(i<n)new=((unsigned int)buf[i].wrd_id)>>8;
			else	new=prev+1;

			if((i==n)||((((unsigned int)prev)>>12)!=(((unsigned int)new)>>12))){
				if(first!=i){
					int newfd,oldfd;
					char scrc[16];
					char dname[UDMSTRSIZ]="";
					char tmpname[UDMSTRSIZ]="";
					int r,w,t,t_prev,t_new,pos,count=0;
					UDM_CACHEHEADER	header;
					UDM_CACHETABLE	table[4096];
					UDM_CACHEWORD cache[4096];
					UDM_LOGWORD    * logwords = NULL;
					int add;

					sprintf(scrc,"%08X",(((unsigned int)prev)|(((unsigned int)log)<<24))&0xFFFFF000);
					sprintf(dname,"%s%c%s%c%02X%c%c%c",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,log,UDMSLASH,scrc[2],UDMSLASH);
					UdmBuild(dname,0755);
					sprintf(dname,"%s%c%s%c%02X%c%c%c",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,log,UDMSLASH,scrc[2],UDMSLASH);

					//sprintf(fname,"%s%c%c",dname,scrc[3],scrc[4]);
					sprintf(fname,"%s%s",dname,scrc);
					printf("%s new: %08X prev: %08X first: %d i: %d\n",fname,new,prev,first,i);

					done[(prev>>12)&0x00000FFF]=1;

					/* Read OLD words from tree */
					if((oldfd=open(fname,O_RDONLY))>=0){
						int j;
						
						printf("Read old: %s\n",fname);
						read(oldfd,&header,sizeof(header));
						read(oldfd,table,header.ntables*sizeof(UDM_CACHETABLE));
							
						for(w=0;w<header.ntables;w++){
							int c=0;
							int n=table[w].len/sizeof(UDM_CACHEWORD);
							while((r=(n-c))>0){
								if(r>4096)r=4096;
								read(oldfd,cache,r*sizeof(UDM_CACHEWORD));
								if(!count)logwords=(UDM_LOGWORD*)malloc((count+r)*sizeof(UDM_LOGWORD));
								else	logwords=(UDM_LOGWORD*)realloc(logwords,(count+r)*sizeof(UDM_LOGWORD));
								for(j=0;j<r;j++){
									/* Add only those OLD words which */
									/* do not present in DEL log      */
									if(!PresentInLog(del_buf,del_count,cache[j].url_id)){
										logwords[count+j].wrd_id=table[w].wrd_id;
										logwords[count+j].url_id=cache[j].url_id;
										logwords[count+j].category=cache[j].category;
										add++;
									}else{
										printf("URL %d was not added\n",cache[j].url_id);
									}
								}
								c+=r;
								count+=add;
							}
						}
						close(oldfd);
					}
					r=i-first;
					
					printf("OLD WORDS: %d\n",add);
					printf("NEW WORDS: %d\n",r);
					
					/* Merge OLD and NEW words */
					if(!count)logwords=(UDM_LOGWORD*)malloc((count+r)*sizeof(UDM_LOGWORD));
					else	logwords=(UDM_LOGWORD*)realloc(logwords,(count+r)*sizeof(UDM_LOGWORD));
					memcpy(logwords+count,buf+first,r*sizeof(UDM_LOGWORD));
					count+=r;
					printf("TOT WORDS: %d\n",count);
					
					/* Sort them in (wrd_id,url_id) order */
					qsort(logwords,count,sizeof(UDM_LOGWORD),cmpcache);

					/* Create table */
					header.version=0;
					header.ntables=0;
					pos=0;
					t_prev=logwords[0].wrd_id;
					for(t=1;t<count+1;t++){
						if(t<count)t_new=logwords[t].wrd_id;
						else	t_new=t_prev+1;
						if(t_prev!=t_new){
							table[header.ntables].wrd_id=t_prev;
							table[header.ntables].pos=pos;
							table[header.ntables].len=t*sizeof(UDM_CACHEWORD)-pos;
							pos+=table[header.ntables].len;
							header.ntables++;
						}
						t_prev=t_new;
					}
					
					/* Write filtered and merged       */
					/* OLD and NEW words into tmp file */
					sprintf(tmpname,"%s%c%c.tmp",dname,scrc[3],scrc[4]);
					newfd=open(tmpname,O_WRONLY|O_CREAT|O_TRUNC,0644);
					write(newfd,&header,sizeof(header));
					write(newfd,table,header.ntables*sizeof(UDM_CACHETABLE));
					
					w=0;
					while(w<count){
						int j;
						if((count-w)>4096)r=4096;
						else	r=(count-w);
						
						for(j=0;j<r;j++){
							cache[j].url_id=logwords[w+j].url_id;
							cache[j].category=logwords[w+j].category;
						}
						w+=r;
						write(newfd,cache,r*sizeof(UDM_CACHEWORD));
					}
					close(newfd);
					UDM_FREE(logwords);
					
					/* rename tmp file into */
					/* real name            */
					rename(tmpname,fname);
				}
				first=i;
			}
			prev=new;
		}
	}
	close(fd);
	UDM_FREE(buf);
	
	/* Remove words from */
	/* undone files      */
	for(i=0;i<4096;i++){
		if(done[i]==0){
			char	n[3];
			sprintf(n,"%03X",i);
			sprintf(dname,"%s%c%s%c%02X%c%c%c",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,log,UDMSLASH,n[0],UDMSLASH);
			sprintf(fname,"%s%02X%s000",dname,log,n);
			UdmDeleteFromCache(fname,del_buf,del_count);
		}
	}
	UDM_FREE(del_buf);
	return IND_OK;
}



/****************** Search stuff ************************************/

typedef struct m {
	int last;
	int cur;
} UDM_MERG;

typedef struct pm {
	UDM_SEARCHWORD * plast;
	UDM_SEARCHWORD * pcur;
} UDM_PMERG;

#define MAXWORD	100
#define MAXMERGE 1024

UDM_SEARCHWORD * UdmFindCache(UDM_AGENT * Indexer,char *q){
	register int i,lex;
	register int nmerg=0;
	register int min_num=0;
	register int nwrd=0;
	register int nwrd1=0;
	register int words_in_query=0;
	register int nitems=0;
	char * w, * lasttok;
	int cat,mask,*url_list;

	UDM_SEARCHWORD * wrd = NULL;
	UDM_SEARCHWORD * wrd1 = NULL;
	char * words[MAXWORD];
	int    cwords[MAXWORD];
	UDM_MERG merg[MAXMERGE];
	UDM_PMERG pmerg[MAXMERGE];
	UDM_STACK_ITEM items[UDM_MAXSTACK];
#ifdef DEBUG_SEARCH
	unsigned long ticks;
#endif


	Indexer->wordinfo[0]='\0';
	Indexer->total_found=0;
	lasttok=q;

	lex=UdmGetLex(&w,&lasttok,Indexer->charset);
	while(lex!=-1){
		char ** ww, *rw;

		if(lex==UDM_STACK_WORD){
			/* FIXME ww are not freed */
			ww=UdmNormalizeWord(Indexer,w);
			rw=ww?*ww:w;
			items[nitems].cmd=UDM_STACK_WORD;
			items[nitems].arg=1<<(words_in_query);
			nitems++;
		}else{
			items[nitems].cmd=lex;
			items[nitems].arg=0;
			nitems++;
			lex=UdmGetLex(&w,&lasttok,Indexer->charset);
			continue;
		}
		if(UdmIsStopWord(Indexer->Conf,rw)){
			if(Indexer->wordinfo[0])strcat(Indexer->wordinfo,", ");
			sprintf(UDM_STREND(Indexer->wordinfo)," %s :stopword",rw);
			lex=UdmGetLex(&w,&lasttok,Indexer->charset);
			continue;
		}
		if(words_in_query<MAXWORD){
			/* FIXME !!! free words[x] */
			words[words_in_query]=strdup(rw);
			cwords[words_in_query]=UdmStrCRC32(rw);
			words_in_query++;
		}
		lex=UdmGetLex(&w,&lasttok,Indexer->charset);
	}
	UdmDecodeCatStr(Indexer->Conf->category,&cat,&mask);
	url_list=UdmGetURLList(Indexer);

#ifdef DEBUG_SEARCH
	fprintf(stderr,"Start read cache %d words\n",words_in_query);
	ticks=UdmStartTimer();
#endif
	for(i=0;i<words_in_query;i++){
		int fd1,t;
		char scrc[16];
		char dname[UDMSTRSIZ];
		char fname[UDMSTRSIZ];
		
		UDM_CACHEHEADER  header;
		UDM_CACHETABLE   table[8192];
		UDM_CACHEWORD  * cw = NULL;
		int data_ofs;
		int group_num=0;
		int group_first=0;
		int group;
		
		sprintf(scrc,"%08X",cwords[i]&0xFFFFF000);
		sprintf(dname,"%s%c%s%c%c%c%c%c%c",
			UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,
			scrc[0],scrc[1],UDMSLASH,scrc[2],UDMSLASH);
#ifdef SHORTNAME
		sprintf(fname,"%s%c%c",dname,scrc[3],scrc[4]);
#else
		sprintf(fname,"%s%s",dname,scrc);
#endif

		if((fd1=open(fname,O_RDONLY))<0){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"\tCan't open: %s\n",fname);
#endif
			continue;
		}
#ifdef DEBUG_SEARCH
	fprintf(stderr,"\tOpened: %s\n",fname);
#endif
		read(fd1,&header,sizeof(header));
		data_ofs=sizeof(header)+sizeof(UDM_CACHETABLE)*header.ntables;
		read(fd1,table,header.ntables*sizeof(UDM_CACHETABLE));
		qsort(&table,header.ntables,sizeof(UDM_CACHETABLE),cmptable);
		
		group_first=nwrd;
		group=0;
		for(t=0;t<header.ntables;t++){
			int bytes;
			if((cwords[i]&0x00FFFFFF)==(((unsigned int)(table[t].wrd_id))>>8)){
				int j,num,real_num=0;
				
				
				cw=(UDM_CACHEWORD*)malloc(table[t].len);
				lseek(fd1,data_ofs+table[t].pos,SEEK_SET);
				bytes=read(fd1,cw,table[t].len);

				if(bytes==-1)fprintf(stderr,"Error reading cache file %s\n",fname);

				num=bytes/sizeof(UDM_CACHEWORD);
				if(!nwrd)wrd=(UDM_SEARCHWORD*)malloc(num*sizeof(UDM_SEARCHWORD));
				else	wrd=(UDM_SEARCHWORD*)realloc(wrd,(num+nwrd)*sizeof(UDM_SEARCHWORD));

				for(j=0;j<num;j++){
					register int add=1;
					
					if(url_list)
						if(!PresentInInt(url_list+1,url_list[0],cw[j].url_id))
							add=0;
					if(((cw[j].category&mask)==cat)&&(add)){
						wrd[nwrd+real_num].url_id=cw[j].url_id;
						if(Indexer->search_mode==UDM_MODE_BOOL){
							wrd[nwrd+real_num].count=1<<i;
						}else{
							wrd[nwrd+real_num].count=1;
						}
						wrd[nwrd+real_num].weight=table[t].wrd_id&0xFF;
						real_num++;
					}
				}
				if(group<2){
					merg[nmerg].cur=nwrd;
					merg[nmerg].last=nwrd+real_num;
					nmerg++;
				}
				group_num+=real_num;
				nwrd+=real_num;
				group++;
			}
		}
		close(fd1);

		if((group==0)&&(Indexer->search_mode==UDM_MODE_ALL)){
			Indexer->total_found=0;
			return(NULL);
		}
		sprintf(UDM_STREND(Indexer->wordinfo)," %s: %d",words[i],group_num);

		if(nmerg){
			merg[nmerg-1].last=nwrd;
			qsort(wrd+merg[nmerg-1].cur,
				merg[nmerg-1].last-merg[nmerg-1].cur,
				sizeof(UDM_SEARCHWORD),cmpurlid);
		}
	}
	
#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"Stop read cache:\t%.2f\n",(float)ticks/1000);
	fprintf(stderr,"Start merge %d groups, %d words\n",nmerg,nwrd);
	ticks=UdmStartTimer();
	fprintf(stderr,"Allocate buffer for merged words\n");
#endif

	if(nwrd){
		/* Allocate memory for merged words */
		wrd1=(UDM_SEARCHWORD*)malloc(nwrd*sizeof(UDM_SEARCHWORD));
		/* Fill pointers array */
		for(i=0;i<nmerg;i++){
			pmerg[i].pcur=wrd+merg[i].cur;
			pmerg[i].plast=wrd+merg[i].last;
		}
	}

	while(nmerg&&nwrd){
		register int k, min_val=0x7FFFFFFF;
		register UDM_PMERG * p;

		if(nmerg==1){
			int num;
			
			num=pmerg[0].plast-pmerg[0].pcur;
			memcpy(&wrd1[nwrd1],pmerg[0].pcur,num*sizeof(UDM_SEARCHWORD));
			nwrd1+=num;
			break;
		}else{
			for(k=0;k<nmerg;k++){
				if(min_val>(pmerg[k].pcur)->url_id){
					min_val=(pmerg[k].pcur)->url_id;
					min_num=k;	
				}
			}
			p=&pmerg[min_num];
			memcpy(&wrd1[nwrd1],p->pcur,sizeof(UDM_SEARCHWORD));
			(p->pcur)++;
			if(p->pcur>=p->plast){
				nmerg--;
				if(min_num<nmerg){
					memmove(p,p+1,sizeof(UDM_PMERG)*(nmerg-min_num));
				}
			}
			nwrd1++;
		}
	}
	UDM_FREE(wrd);

#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"Stop merge groups:\t%.2f nwrd=%d nwrd1=%d\n",(float)ticks/1000,nwrd,nwrd1);
#endif
	
	if((Indexer->total_found=nwrd1)){
		UdmGroupByURL(Indexer,wrd1,words_in_query,items,nitems);
	}
	return(wrd1);
}
