#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_charset.h"
#include "udm_proto.h"
#include "udm_parseurl.h"
#include "udm_parser.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_log.h"
#include "udm_hrefs.h"
#include "udm_robots.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_host.h"

int UdmInit(){
	UdmInitCharset();
	UdmInitTZ();
	return(0);
}



/*******************************************************************/
typedef struct filter_struct {
	regex_t filter;
	int     filter_type;
	int	reverse;
	char	*regstr;
} UDM_FILTER;


/************* Base 64 *********************************************/
/* BASE64 encoding converts  3x8 bits into 4x6 bits                */
/* Macro for length required for "len" bytes of non-encoded string */
#define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)

static void base64_encode (const char *s, char *store, int length){
int i;
unsigned char *p = (unsigned char *)store;
char base64[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

	for (i = 0; i < length; i += 3){
		*p++ = base64[s[0] >> 2];
		*p++ = base64[((s[0] & 3) << 4) + (s[1] >> 4)];
		*p++ = base64[((s[1] & 0xf) << 2) + (s[2] >> 6)];
		*p++ = base64[s[2] & 0x3f];
		s += 3;
	}
	/*Pad the result*/
	if (i == length + 1) *(p - 1) = '=';
	else if (i == length + 2) *(p - 1) = *(p - 2) = '=';
	*p = '\0';
}


/************ Allow/Disallow/CheckOnly  stuff *****************/

static int AddFilter(UDM_ENV * Conf,char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char regerrstr[UDMSTRSIZ]="";
int err;
UDM_FILTER * Filter;
	if(Conf->nfilters>=Conf->mfilters){
		if(Conf->mfilters){
			Conf->mfilters+=16;
			Conf->Filter=(UDM_FILTER *)realloc(Conf->Filter,Conf->mfilters*sizeof(UDM_FILTER));
		}else{
			Conf->mfilters=16;
			Conf->Filter=(UDM_FILTER *)malloc(Conf->mfilters*sizeof(UDM_FILTER));
		}
	}
	Filter=(UDM_FILTER*)Conf->Filter;
	err=regcomp(&(Filter[Conf->nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err, &(Filter[Conf->nfilters].filter), regerrstr, ERRSTRSIZE);
		Conf->errcode=1;
		sprintf(Conf->errstr,"Wrong regex in config file: %s: %s", filter,regerrstr);
		regfree(&(Filter[Conf->nfilters].filter));
		return(1);
	}
	Filter[Conf->nfilters].filter_type=filter_type;
	Filter[Conf->nfilters].reverse=reverse;
	Filter[Conf->nfilters].regstr=strdup(filter);
	Conf->nfilters++;
	return(0);
}
static void FreeFilters(UDM_ENV * Conf){
int i;
UDM_FILTER * Filter;
	Filter=(UDM_FILTER*)Conf->Filter;
	for(i=0; i < Conf->nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	UDM_FREE(Conf->Filter);
	Conf->nfilters=0;
	Conf->mfilters=0;
}
int UdmFilterType(UDM_ENV * Conf,char * param,char * reason){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
UDM_FILTER * Filter;

	Filter=(UDM_FILTER *)Conf->Filter;
	reason[0]=0;
	for(i=0;i<Conf->nfilters;i++){
		err=regexec(&(Filter[i].filter),param,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			break;
		if((!Filter[i].reverse)&&(!err))
			break;
	}
	if(i<Conf->nfilters){
		switch(Filter[i].filter_type){
			case UDM_ALLOW:		strcpy(reason,"Allow");break;
			case UDM_DISALLOW:	strcpy(reason,"Disallow");break;
			case UDM_HEAD:		strcpy(reason,"CheckOnly");break;
			case UDM_HREFONLY:	strcpy(reason,"HrefOnly");break;
			default:		strcpy(reason,"Unknown");break;
		}
		strcpy(UDM_STREND(reason),Filter[i].reverse?"NoMatch":"");
		strcpy(UDM_STREND(reason),Filter[i].regstr?Filter[i].regstr:"");
		return(Filter[i].filter_type);
	}
	strcpy(reason,"Allow by default");
	return(UDM_ALLOW);
}

/******************** ALIASES ***************************************/
static int cmpalias(const void *s1,const void *s2){
	return(strcmp(((UDM_ALIAS*)(s1))->find,((UDM_ALIAS*)(s2))->find));
}
static int AddAlias(UDM_ENV * Conf,char * find,char *replace){
	if(Conf->naliases>=Conf->maliases){
		if(Conf->maliases){
			Conf->maliases+=16;
			Conf->Alias=(UDM_ALIAS *)realloc(Conf->Alias,Conf->maliases*sizeof(UDM_ALIAS));
		}else{
			Conf->maliases=16;
			Conf->Alias=(UDM_ALIAS *)malloc(Conf->maliases*sizeof(UDM_ALIAS));
		}
	}
	Conf->Alias[Conf->naliases].find=strdup(find);
	Conf->Alias[Conf->naliases].replace=strdup(replace);
	Conf->naliases++;
	return(0);
}
static void FreeAliases(UDM_ENV * Conf){
int i;
	for(i=0;i<Conf->naliases;i++){
		free(Conf->Alias[i].find);
		free(Conf->Alias[i].replace);
	}
	Conf->naliases=0;
	Conf->maliases=0;
	UDM_FREE(Conf->Alias);
}
UDM_ALIAS * UdmFindAlias(UDM_ENV * Conf,char * alias){
	register int l,c,r,res;
	
	l=0;r=Conf->naliases-1;
	while(l<=r){
		c=(l+r)>>1;
		res=strncmp(Conf->Alias[c].find,alias,strlen(Conf->Alias[c].find));
		if(res<0){
			l=c+1;
		}else
		if(res>0){
			r=c-1;	
		}else{
			return(&Conf->Alias[c]);
		}
	}
	return(NULL);
}
/*************************** Servers *************************************/

#define DEFAULT_PROXY_PORT	3128

static int AddServer(UDM_ENV * Conf,UDM_SERVER * srv){
	if(Conf->nservers>=Conf->mservers){
		if(Conf->mservers){
			Conf->mservers+=16;
			Conf->Server=(UDM_SERVER *)realloc(Conf->Server,Conf->mservers*sizeof(UDM_SERVER));
		}else{
			Conf->mservers=16;
			Conf->Server=(UDM_SERVER *)malloc(Conf->mservers*sizeof(UDM_SERVER));
		}
	}
	Conf->Server[Conf->nservers].url=strdup(srv->url);
	Conf->Server[Conf->nservers].charset=srv->charset?strdup(srv->charset):NULL;
	Conf->Server[Conf->nservers].basic_auth=srv->basic_auth?strdup(srv->basic_auth):NULL;
	Conf->Server[Conf->nservers].proxy_basic_auth=srv->proxy_basic_auth?strdup(srv->proxy_basic_auth):NULL;
	Conf->Server[Conf->nservers].htdb_list=srv->htdb_list?strdup(srv->htdb_list):NULL;
	Conf->Server[Conf->nservers].htdb_doc=srv->htdb_doc?strdup(srv->htdb_doc):NULL;
        Conf->Server[Conf->nservers].user=srv->user?strdup(srv->user):NULL;
        Conf->Server[Conf->nservers].passwd=srv->passwd?strdup(srv->passwd):NULL;
        Conf->Server[Conf->nservers].category=srv->category?strdup(srv->category):NULL;
        Conf->Server[Conf->nservers].tag=srv->tag?strdup(srv->tag):NULL;
        Conf->Server[Conf->nservers].mirror_root=srv->mirror_root?strdup(srv->mirror_root):NULL;
        Conf->Server[Conf->nservers].mirror_headers=srv->mirror_headers?strdup(srv->mirror_headers):NULL;
	if(srv->proxy){
		char *s;
		Conf->Server[Conf->nservers].proxy=strdup(srv->proxy);
		if((s=strchr(Conf->Server[Conf->nservers].proxy,':'))){
			*s=0;
			Conf->Server[Conf->nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Conf->Server[Conf->nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Conf->Server[Conf->nservers].proxy=NULL;
		Conf->Server[Conf->nservers].proxy_port=0;
	}
	Conf->Server[Conf->nservers].period=srv->period;
	Conf->Server[Conf->nservers].net_errors=0;
	Conf->Server[Conf->nservers].outside=srv->outside;
	Conf->Server[Conf->nservers].maxhops=srv->maxhops;
	Conf->Server[Conf->nservers].gindex=srv->gindex;
	Conf->Server[Conf->nservers].gfollow=srv->gfollow;
	Conf->Server[Conf->nservers].deletebad=srv->deletebad;
	Conf->Server[Conf->nservers].userobots=srv->userobots;
	Conf->Server[Conf->nservers].bodyweight=srv->bodyweight;
	Conf->Server[Conf->nservers].titleweight=srv->titleweight;
	Conf->Server[Conf->nservers].urlweight=srv->urlweight;
	Conf->Server[Conf->nservers].urlhostweight=srv->urlhostweight;
	Conf->Server[Conf->nservers].urlpathweight=srv->urlpathweight;
	Conf->Server[Conf->nservers].urlfileweight=srv->urlfileweight;
	Conf->Server[Conf->nservers].descweight=srv->descweight;
	Conf->Server[Conf->nservers].keywordweight=srv->keywordweight;
	Conf->Server[Conf->nservers].max_net_errors=srv->max_net_errors;
	Conf->Server[Conf->nservers].read_timeout=srv->read_timeout;
	Conf->Server[Conf->nservers].delete_no_server=srv->delete_no_server;
	Conf->Server[Conf->nservers].correct_factor=srv->correct_factor;
	Conf->Server[Conf->nservers].incorrect_factor=srv->incorrect_factor;
	Conf->Server[Conf->nservers].number_factor=srv->number_factor;
	Conf->Server[Conf->nservers].alnum_factor=srv->alnum_factor;
	Conf->Server[Conf->nservers].min_word_length=srv->min_word_length;
	Conf->Server[Conf->nservers].max_word_length=srv->max_word_length;
	Conf->Server[Conf->nservers].use_mirror=srv->use_mirror;
	Conf->Server[Conf->nservers].use_clones=srv->use_clones;
	Conf->Server[Conf->nservers].check_mp3_tag=srv->check_mp3_tag;
	Conf->Server[Conf->nservers].check_only_mp3_tag=srv->check_only_mp3_tag;
	Conf->Server[Conf->nservers].net_error_delay_time=srv->net_error_delay_time;
	Conf->nservers++;
	return(0);
}

static void FreeServers(UDM_ENV * Conf){
int i;
	for(i=0;i<Conf->nservers;i++){
		UDM_FREE(Conf->Server[i].url);
		UDM_FREE(Conf->Server[i].charset);
		UDM_FREE(Conf->Server[i].basic_auth);
		UDM_FREE(Conf->Server[i].proxy_basic_auth);
		UDM_FREE(Conf->Server[i].htdb_list);
		UDM_FREE(Conf->Server[i].htdb_doc);
		UDM_FREE(Conf->Server[i].proxy);
		UDM_FREE(Conf->Server[i].category);
		UDM_FREE(Conf->Server[i].tag);
		UDM_FREE(Conf->Server[i].mirror_root);
		UDM_FREE(Conf->Server[i].mirror_headers);
	}
	Conf->nservers=Conf->mservers=0;
	UDM_FREE(Conf->Server);
}

UDM_SERVER * UdmFindServer(UDM_ENV * Conf,char *url){
int i;
char *r;

	if(!Conf->Server)return(NULL);
	url=UdmRemove2Dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<Conf->nservers;i++){
			if(!UDM_STRNCMP(url,Conf->Server[i].url)){
				*r='r';
				return(&Conf->Server[i]);
			}
		}
		*r='r';
	}else{
		for(i=0;i<Conf->nservers;i++){
			if(!UDM_STRNCMP(url,Conf->Server[i].url))
				return(&Conf->Server[i]);
		}
	}
	return(0);
}
static int cmpserver(const void *s1,const void *s2){
	return(strlen(((UDM_SERVER*)s2)->url)-strlen(((UDM_SERVER*)s1)->url));
}
static int InitServer(UDM_SERVER * srv){
	srv->url=NULL;
	srv->proxy=NULL;
	srv->basic_auth=NULL;
	srv->proxy_basic_auth=NULL;
	srv->charset=UDM_CHARSET_USASCII;
	srv->htdb_list=NULL;
	srv->htdb_doc=NULL;
	srv->category=NULL;
	srv->proxy_port=0;
	srv->period=UDM_DEFAULT_REINDEX_TIME;
	srv->tag=NULL;
	srv->mirror_root=NULL;
	srv->mirror_headers=NULL;
	srv->net_errors=0;
	srv->max_net_errors=UDM_MAXNETERRORS;
	srv->read_timeout=UDM_READ_TIMEOUT;
	srv->outside=0;
	srv->delete_no_server=1;
	srv->maxhops=UDM_DEFAULT_MAX_HOPS;
	srv->gindex=1;
	srv->gfollow=1;
	srv->deletebad=0;
	srv->userobots=1;
	srv->bodyweight=1;
	srv->titleweight=2;
	srv->keywordweight=2;
	srv->descweight=2;
	srv->urlweight=0;
	srv->urlhostweight=0;
	srv->urlpathweight=0;
	srv->urlfileweight=0;
	srv->correct_factor=1;
	srv->incorrect_factor=1;
	srv->number_factor=1;
	srv->alnum_factor=1;
	srv->min_word_length=1;
	srv->max_word_length=32;
	srv->use_mirror=UDM_MIRROR_NO;
	srv->use_clones=1;
	srv->net_error_delay_time=UDM_DEFAULT_NET_ERROR_DELAY_TIME;
	
	return(0);
}

/**************************** DBAddr ***********************************/
static int UdmStr2DBMode(char * str1){
int m;
	m = UDM_DBMODE_SINGLE;
	if(!UDM_STRNCASECMP(str1,"multi-crc"))m=UDM_DBMODE_MULTI_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"crc-multi"))m=UDM_DBMODE_MULTI_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"single"))m=UDM_DBMODE_SINGLE;
	else
	if(!UDM_STRNCASECMP(str1,"crc"))m=UDM_DBMODE_SINGLE_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"multi"))m=UDM_DBMODE_MULTI;
	else
	if(!UDM_STRNCASECMP(str1,"word2url"))m=UDM_DBMODE_WORD2URL;
	else
	if(!UDM_STRNCASECMP(str1,"cache"))m=UDM_DBMODE_CACHE;
	return(m);
}
__INDLIB__ int UdmEnvSetDBMode(UDM_ENV * Conf,char *str){
	Conf->DBMode=UdmStr2DBMode(str);
	return(Conf->DBMode);
}
static int UdmStr2DBType(char *str1) {
int t=UDM_DB_UNK;
	if(!UDM_STRNCASECMP(str1,"msql"))t=UDM_DB_MSQL;
	else
	if(!UDM_STRNCASECMP(str1,"solid"))t=UDM_DB_SOLID;
	else
	if(!UDM_STRNCASECMP(str1,"oracle"))t=UDM_DB_ORACLE;
	else
	if(!UDM_STRNCASECMP(str1,"mssql"))t=UDM_DB_MSSQL;
	else
	if(!UDM_STRNCASECMP(str1,"mysql"))t=UDM_DB_MYSQL;
	else
	if(!UDM_STRNCASECMP(str1,"pgsql"))t=UDM_DB_PGSQL;
	else
	if(!UDM_STRNCASECMP(str1,"ibase"))t=UDM_DB_IBASE;
	return(t);
}
__INDLIB__ int UdmEnvSetDBAddr(UDM_ENV * Conf,char * addr){
	UDM_URL url;
	char buf[1024];
	char *s;
	int res;

	strncpy(buf,addr,sizeof(buf));
	buf[sizeof(buf)-1]=0;

	if((res=UdmParseURL(&url,buf)))
		return(res);

	Conf->DBHost=strdup(url.hostname);
	Conf->DBType=UdmStr2DBType(url.schema);

	if(Conf->DBType==UDM_DB_IBASE){
		/* Ibase is a special case        */
		/* It's database name consists of */
		/* full path and file name        */ 
		Conf->DBName=malloc(strlen(url.path)+1+strlen(url.filename));
		sprintf(Conf->DBName,"%s%s",url.path,url.filename);
	}else{
		Conf->DBName=strdup(url.path);
		sscanf(url.path,"/%[^/]s",Conf->DBName);
	}
	Conf->DBPort=url.port;
	if((s=strchr(url.auth,':'))){
		*s=0;
		Conf->DBUser=strdup(url.auth);
		Conf->DBPass=strdup(s+1);
		*s=':';
	}else{
		Conf->DBUser=strdup(url.auth);
	}
	return(0);
}

/****************************  Load Configuration **********************/


#define IF_NUM_PARAM(src,name,dst,def) \
if(!UDM_STRNCASECMP(src,name)){if(sscanf(src+strlen(name),"%d",&(dst))!=1)dst=def;}

#define IF_BOOL_PARAM(src,name,dst,def) \
if(!UDM_STRNCASECMP(src,name)){ \
char xxx[1024]; \
if(sscanf(src+strlen(name),"%s",xxx)==1){ \
	if(!UDM_STRNCASECMP(xxx,"yes"))dst=1; \
	else	dst=0; \
}else \
	dst=def; \
}


__INDLIB__  UDM_ENV * UdmLoadConfig(UDM_ENV * Conf,char *conf_name,int config_level,int load_flags){
int i=0; /* For line number */
FILE *config;
char str[UDMSTRSIZ]="";
char str1[UDMSTRSIZ]="";
char str2[UDMSTRSIZ]="";
char *ch;
int has_dict=0;

	Conf->errcode=0;
	Conf->errstr[0]=0;
	Conf->force1251=0;

	if(config_level==0){ /* Do some initialization */
		Conf->csrv=(UDM_SERVER*)UdmXmalloc(sizeof(UDM_SERVER));
		InitServer(Conf->csrv);
		sprintf(Conf->user_agent,"%s/%s",UDM_USER_AGENT,VERSION);
		Conf->extra_headers[0]=0;
		Conf->max_doc_size=UDM_MAXDOCSIZE;
		UdmSetLogFacility(Conf, "");
		Conf->local_charset=UDM_CHARSET_USASCII;

		/* I have removed it to allow adding of URLs  */
		/* Passed in -i with -u command line args     */
		/* But it seems that UnloadConf() is required */
		/* REMOVED      UdmFreeHrefs(); */

		
	}

	/* Open config file */
	if(!(config=fopen(conf_name,"r"))){
		sprintf(Conf->errstr,"Error: can't open config file '%s': %s",conf_name, strerror(errno));
		Conf->errcode=1;
		return(Conf);
	}

	/*  Read lines and parse */
	while(fgets(str1,sizeof(str1),config)){
		char *end;
		i++;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		if(!UDM_STRNCASECMP(str,"DBAddr")){
			if(UdmEnvSetDBAddr(Conf,UdmTrim(str+6," \r\n\t"))){
				sprintf(Conf->errstr,"Bad DBAddr in config file '%s' line %d:%s",conf_name,i,str);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"DBMode")){
			Conf->DBMode=UdmStr2DBMode(UdmTrim(str+6," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"LocalCharset")){
			Conf->local_charset=UdmGetCharset(UdmTrim(str+13," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"DisallowNoMatch")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_DISALLOW,1))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"Disallow")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_DISALLOW,0))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"AllowNoMatch")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_ALLOW,1))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"Allow")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_ALLOW,0))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"CheckOnlyNoMatch")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_HEAD,1))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"CheckOnly")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_HEAD,0))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"HrefOnly")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_HREFONLY,0))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"HrefOnlyNoMatch")){
			char *s,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			while((s=UdmGetToken(NULL," \t\r\n",&lt))){
				if(AddFilter(Conf,s,UDM_HREFONLY,1))
					return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"AddType")){
			char *s,*s1,*lt;
			s=UdmGetToken(str," \t\r\n",&lt);
			if((s1=UdmGetToken(NULL," \t\r\n",&lt)))
				while((s=UdmGetToken(NULL," \t\r\n",&lt)))
					UdmAddType(Conf,s1,s,Conf->errstr);
		}else
		if(!UDM_STRNCASECMP(str,"CharSet")){
			UDM_FREE(Conf->csrv->charset);
			if(sscanf(str+7,"%s",str1))
				Conf->csrv->charset=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"ProxyAuthBasic")){
			UDM_FREE(Conf->csrv->proxy_basic_auth);
			if(sscanf(str+14,"%s",str1)){
				Conf->csrv->proxy_basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,Conf->csrv->proxy_basic_auth,strlen(str1));
			}
		}else
		if(!UDM_STRNCASECMP(str,"Proxy")){
			UDM_FREE(Conf->csrv->proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				Conf->csrv->proxy=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"Category")){
			UDM_FREE(Conf->csrv->category);
			if(sscanf(str+9,"%s",str1))
			Conf->csrv->category=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"Tag")){
			UDM_FREE(Conf->csrv->tag);
			if(sscanf(str+3,"%s",str1))
			Conf->csrv->tag=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"HTTPHeader")){
			if(sscanf(str+11,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!UDM_STRNCMP(str1,"User-Agent: ")){
					strcpy(Conf->user_agent,str1+12);
				}else{
					strcat(Conf->extra_headers,str1);
					strcat(Conf->extra_headers,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H
		if(!UDM_STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				UdmSetLogFacility(Conf, str1);
		}else
#endif
		if(!UDM_STRNCASECMP(str,"AuthBasic")){
			UDM_FREE(Conf->csrv->basic_auth);
			UDM_FREE(Conf->csrv->user);
			UDM_FREE(Conf->csrv->passwd);
			if(sscanf(str+9,"%s",str1)){
				Conf->csrv->basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,Conf->csrv->basic_auth,strlen(str1));
                                if ((ch=strchr(str1, ':'))){
                            		Conf->csrv->user=(char*)UdmXmalloc(ch-str1+1);
                            		udm_snprintf(Conf->csrv->user, ch-str1+1, "%s", str1);
                            		Conf->csrv->passwd=(char*)UdmXmalloc(strlen(ch)+1);
                            		sprintf(Conf->csrv->passwd, "%s", ch+1);
                                }

			}
		}else
		if(!UDM_STRNCASECMP(str,"HTDBList")){
			UDM_FREE(Conf->csrv->htdb_list);
			Conf->csrv->htdb_list=strdup(UdmTrim(str+8," \t\r\n"));
		}else
		if(!UDM_STRNCASECMP(str,"HTDBDoc")){
			UDM_FREE(Conf->csrv->htdb_doc);
			Conf->csrv->htdb_doc=strdup(UdmTrim(str+8," \t\r\n"));
		}else
		if(!UDM_STRNCASECMP(str,"Mime")){
			UdmAddParser(Conf,str+4);
		}else
		if(!UDM_STRNCASECMP(str,"Alias")){
			if(sscanf(str+5,"%s%s",str1,str2)==2)
				AddAlias(Conf,str1,str2);
			else{
				sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"Server")){
			UDM_URL from;
			int res;
			if(sscanf(str+6,"%s",str1)!=1)
				continue;
			if((res=UdmParseURL(&from,str1))){
				switch(res){
				case UDM_PARSEURL_LONG:
					Conf->errcode=1;
					sprintf(Conf->errstr,"File '%s' line %d: URL too long: %s",conf_name,i,str1);
					break;
				case UDM_PARSEURL_BAD:
				default:
					Conf->errcode=1;
					sprintf(Conf->errstr,"File '%s' line %d: Bad formed URL: %s",conf_name,i,str1);
				}
				return(Conf);
			}
			Conf->csrv->url=strdup(str1);
			if (!strcmp(from.schema, "news")) {
				char *c, *cc;
				/* Cat server name to remove group names */
				/* This is because group names do not    */
				/* present in message URL                */
				c=Conf->csrv->url+7;
				cc=strchr(c,'/');
				if(cc)*(cc+1)='\0';
			}
			AddServer(Conf,Conf->csrv);
			if((!strcmp(from.schema,"http"))&&(load_flags&UDM_FLAG_ADD_SERV)&&(Conf->csrv->userobots)){
				sprintf(str,"%s://%s/%s",from.schema,from.hostinfo,"robots.txt");
				UdmAddHref(Conf,str,0,0,0);
			}
			if(load_flags&UDM_FLAG_ADD_SERV)
				UdmAddHref(Conf,str1,0,0,0);
		}
		else	IF_BOOL_PARAM(str,"FollowOutside",Conf->csrv->outside,0)
		else	IF_BOOL_PARAM(str,"Index",Conf->csrv->gindex,1)
		else	IF_BOOL_PARAM(str,"Follow",Conf->csrv->gfollow,1)
		else	IF_BOOL_PARAM(str,"Robots",Conf->csrv->userobots,1)
		else	IF_BOOL_PARAM(str,"DeleteBad",Conf->csrv->deletebad,0)
		else	IF_BOOL_PARAM(str,"DeleteNoServer",Conf->csrv->delete_no_server,1)
		else	IF_BOOL_PARAM(str,"ForceIISCharset1251",Conf->force1251,0)
		else	IF_BOOL_PARAM(str,"Clones",Conf->csrv->use_clones,1)
		else	IF_BOOL_PARAM(str,"CheckMp3Tag",Conf->csrv->check_mp3_tag,0)
		else	IF_BOOL_PARAM(str,"CheckID3TagOnly",Conf->csrv->check_only_mp3_tag,0)
		
		else	IF_NUM_PARAM(str,"MaxNetErrors",Conf->csrv->max_net_errors,UDM_MAXNETERRORS)
		else	IF_NUM_PARAM(str,"NetErrorDelayTime",Conf->csrv->net_error_delay_time,UDM_DEFAULT_NET_ERROR_DELAY_TIME)
		else	IF_NUM_PARAM(str,"ReadTimeOut",Conf->csrv->read_timeout,UDM_READ_TIMEOUT)
		else	IF_NUM_PARAM(str,"Period",Conf->csrv->period,UDM_DEFAULT_REINDEX_TIME)
		else	IF_NUM_PARAM(str,"MaxHops",Conf->csrv->maxhops,0)
		else	IF_NUM_PARAM(str,"BodyWeight",Conf->csrv->bodyweight,1)
		else	IF_NUM_PARAM(str,"TitleWeight",Conf->csrv->titleweight,2)
		else	IF_NUM_PARAM(str,"DescWeight",Conf->csrv->descweight,2)
		else	IF_NUM_PARAM(str,"UrlWeight",Conf->csrv->urlweight,0)
		else    IF_NUM_PARAM(str,"UrlHostWeight",Conf->csrv->urlhostweight,0)
		else    IF_NUM_PARAM(str,"UrlPathWeight",Conf->csrv->urlpathweight,0)
		else    IF_NUM_PARAM(str,"UrlFileWeight",Conf->csrv->urlfileweight,0)
		else	IF_NUM_PARAM(str,"IspellCorrectFactor",Conf->csrv->correct_factor,1)
		else	IF_NUM_PARAM(str,"IspellIncorrectFactor",Conf->csrv->incorrect_factor,1)
		else	IF_NUM_PARAM(str,"NumberFactor",Conf->csrv->number_factor,1)
		else	IF_NUM_PARAM(str,"AlnumFactor",Conf->csrv->alnum_factor,1)
		else	IF_NUM_PARAM(str,"MinWordLength",Conf->csrv->min_word_length,1)
		else	IF_NUM_PARAM(str,"MaxWordLength",Conf->csrv->max_word_length,32)
		else	IF_NUM_PARAM(str,"KeywordWeight",Conf->csrv->keywordweight,2)
		else	IF_NUM_PARAM(str,"MaxDocSize",Conf->max_doc_size,UDM_MAXDOCSIZE)
		else	if(!UDM_STRNCASECMP(str,"NoIndex"))Conf->csrv->gindex=0;
		else	if(!UDM_STRNCASECMP(str,"NoFollow"))Conf->csrv->gfollow=0;
		else
		if(!UDM_STRNCASECMP(str,"Affix")){
			char lang[20];
			if(load_flags&&UDM_FLAG_SPELL){
				if(2==sscanf(str+5,"%s%s",lang,str1)){
					if(*str1=='/')strcpy(str,str1);
					else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
					if(UdmImportAffixes(Conf,lang,str,NULL,0)){
						Conf->errcode=1;
						sprintf(Conf->errstr,"Can't load affix :%s",str);
						return(Conf);
					}
				}
				else {
					Conf->errcode=1;
					sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Spell")){
			char lang[20];
			if(load_flags&&UDM_FLAG_SPELL){
				if(2==sscanf(str+5,"%s%s",lang,str1)){
					if(*str1=='/')strcpy(str,str1);
					else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
					/* -----  0 or 1 ?  */
					if(UdmImportDictionary(Conf,lang,str,0,"")){
						Conf->errcode=1;
						sprintf(Conf->errstr,"Can't load dictionary :%s",str);
						return(Conf);
					}
					has_dict=1;
				}else{
					Conf->errcode=1;
					sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
				UdmLoadConfig(Conf,str,config_level+1,load_flags);
			}
		}else
		if(!UDM_STRNCASECMP(str,"MirrorRoot")){
			if((sscanf(str+10,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
			} else {
				sprintf(str,"%s/mirrors",UDM_CONF_DIR);
			}
			Conf->csrv->mirror_root = strdup(str);
			Conf->csrv->use_mirror=(Conf->csrv->use_mirror>0)?Conf->csrv->use_mirror:0;
		}else
		if(!UDM_STRNCASECMP(str,"MirrorHeadersRoot")){
			if((sscanf(str+17,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
			} else {
				sprintf(str,"%s/headers",UDM_CONF_DIR);
			}
			Conf->csrv->mirror_headers = strdup(str);
			Conf->csrv->use_mirror=(Conf->csrv->use_mirror>0)?Conf->csrv->use_mirror:0;
		}else
		IF_NUM_PARAM(str,"MirrorPeriod",Conf->csrv->use_mirror,UDM_MIRROR_NO)
		else{
			Conf->errcode=1;
			sprintf(Conf->errstr,"Error in config file line %d: %s",i,str);
			return(Conf);
		}
		str[0]=0;
	}
	fclose(config);

	/* On level0 : Free some variables, prepare others, etc */
	if(config_level==0){
		/* Add one virtual server if we want FollowOutside */
		/* or DeleteNoServer no	*/
		
		if((Conf->csrv->outside)||(!Conf->csrv->delete_no_server)){
			UDM_FREE(Conf->csrv->url);
			Conf->csrv->url="";
			AddServer(Conf,Conf->csrv);
		}else{
			UDM_FREE(Conf->csrv->url);
		}
		UDM_FREE(Conf->csrv->charset);
		UDM_FREE(Conf->csrv->proxy);
		UDM_FREE(Conf->csrv->basic_auth);
		UDM_FREE(Conf->csrv->proxy_basic_auth);
		UDM_FREE(Conf->csrv->htdb_list);
		UDM_FREE(Conf->csrv->htdb_doc);
		/*  Long name should be found first    */
		/*  to allow different options         */
		/*  for server and it's subdirectories */
		qsort((void*)Conf->Server,Conf->nservers,sizeof(UDM_SERVER),cmpserver);
		/* Sort Aliases for binary search */
		qsort((void*)Conf->Alias,Conf->naliases,sizeof(UDM_ALIAS),cmpalias);
		/* Sort ispell dictionay if it has been loaded */
		if(has_dict)UdmSortDictionary(Conf);
	}
	return(0);
}

/******************************************************/

__INDLIB__ UDM_ENV * UdmAllocEnv(void){
	UDM_ENV * Conf;

	Conf=(UDM_ENV *)malloc(sizeof(UDM_ENV));
	Conf->errcode=0;
	Conf->errstr[0]=0;
	Conf->local_charset=UDM_CHARSET_USASCII;
	Conf->force1251=0;
	Conf->max_doc_size=UDM_MAXDOCSIZE;
	Conf->user_agent[0]=0;
	Conf->extra_headers[0]=0;
	Conf->url_file_name=NULL;
	Conf->nservers=0;
	Conf->mservers=0;
	Conf->Server=NULL;
	Conf->naliases=0;
	Conf->maliases=0;
	Conf->Alias=NULL;
	Conf->mhrefs=0;
	Conf->nhrefs=0;
	Conf->shrefs=0;
	Conf->dhrefs=0;
	Conf->Href=NULL;	            
	Conf->have_targets=1;
	Conf->nrows=0;
	Conf->currow=0;
	Conf->Filter=NULL;
	Conf->nfilters=0;
	Conf->mfilters=0;	
	Conf->Mime=NULL;
	Conf->nmimes=0;
	Conf->nmimes=0;
	Conf->nparsers=0;
	/*Conf->parsers=NULL;  FIXME */
	Conf->nhost_addr=0;
	Conf->mhost_addr=0;
	Conf->host_addr=NULL;
	Conf->naffixes=0;
	Conf->maffixes=0;
	Conf->Affix=NULL;
	Conf->nspell=0;
	Conf->mspell=0;
	Conf->Spell=NULL;
	Conf->nstoplist=0;
	Conf->stoplist=NULL;
	Conf->ispell_mode=UDM_ISPELL_MODE_TEXT;
	Conf->Robots=NULL;
	Conf->DBHost=NULL;
	Conf->DBName=NULL;
	Conf->DBUser=NULL;
	Conf->DBPass=NULL;
	Conf->DBPort=0;
	Conf->DBType=UDM_DB_UNK;
	Conf->DBMode=UDM_DBMODE_SINGLE;
	Conf->DBUseLock=1;
	Conf->catstr[0]=0;
	Conf->langstr[0]=0;
	Conf->tagstr[0]=0;
	Conf->statusstr[0]=0;
	Conf->urlstr[0]=0;
	Conf->timestr[0]=0;
	Conf->category[0]=0;
	Conf->ul=NULL;
	Conf->ttag=NULL;
	Conf->n_urls=0;
	Conf->URLs=0;
	Conf->urlres=0;
	Conf->ThreadInfo=NULL;
	Conf->StatInfo=NULL;
	Conf->RefInfo=NULL;
	Conf->LockProc=NULL;
	/* for logger */
	/* Set default facility & loglevel */                                       
	Conf->logFacility=LOG_FACILITY;                                              
	Conf->logLevel=UDM_LOG_INFO;                                                 
	Conf->is_log_open=0;
	Conf->logFD=(FILE *)NULL;
	return(Conf);
}
__INDLIB__ void UdmFreeEnv(UDM_ENV * Env){
	if(Env){
		UDM_FREE(Env->DBHost);
		UDM_FREE(Env->DBName);
		UDM_FREE(Env->DBUser);
		UDM_FREE(Env->DBPass);
		UDM_FREE(Env->csrv);

		UdmFreeParsers(Env);
		UdmFreeStopList(Env);
		UdmFreeRobots(Env);
		UdmFreeHost_addr(Env);
		UdmFreeTypes(Env);
		UdmFreeHrefs(Env);
		FreeServers(Env);
		FreeFilters(Env);
		FreeAliases(Env);
		FreeServers(Env);
		free(Env);
	}
}
__INDLIB__ int UdmEnvErrCode(UDM_ENV * Conf){
	return(Conf->errcode);
}
__INDLIB__ char * UdmEnvErrMsg(UDM_ENV * Conf){
	return(Conf->errstr);
}
