/*   implementation of the own backend 
 *   to use UdmSearch without third-party libraries
 */

#include "udm_config.h"

#ifdef HAVE_FILES
#include <stdio.h>
#if (WIN32|WINNT)
#else
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_charset.h"
#include "udm_robots.h"
#include "udm_mutex.h"
#include "udm_utils.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_crc32.h"


static const char udmver[]= VERSION "/UdmDB";
__INDLIB__ const char * UdmVersion(){
	return(udmver);
}


char * UdmDBEscStr(int DBType,char *x,char *y){
	return(strcpy(x,y));
}


/************************* Backen functions *****************/
#define MAXMULTI	32
typedef struct txt_db {
	int connected;
	FILE * dict;
	FILE * url;
	FILE * robots;
	FILE * stopwords;
	int  crcdict[MAXMULTI];
	int  open_mode;
	int  errcode;
	char errmsg[1024];
} DB;

static void CloseDB(void * vdb){
DB * db;
int i;
	db=(DB*)vdb;
	if(db->dict)fclose(db->dict);
	if(db->url)fclose(db->url);
	if(db->robots)fclose(db->robots);
	if(db->stopwords)fclose(db->stopwords);
	for(i=0;i<MAXMULTI;i++)
		if(db->crcdict[i]>=0)
			close(db->crcdict[i]);
}

void UdmFreeDB(UDM_AGENT * Indexer){
        if(Indexer->db){                                                        
                UdmCloseCache(Indexer);                                         
                CloseDB((DB*)(Indexer->db));                                    
                free(Indexer->db);                                              
        }                                                                       
}

static int InitDB(UDM_AGENT * Indexer){
char * fmode;
int  flags;
int  i;
DB * db=Indexer->db;
	if(db->connected)
		return(0);
	else
		db->connected=1;

	for(i=0;i<MAXMULTI;i++)
		db->crcdict[i]=-1;
	switch(db->open_mode){
		case UDM_OPEN_MODE_READ:
			fmode="r";
			flags=O_RDONLY;
			break;
		case UDM_OPEN_MODE_WRITE:
		default:
			fmode="w";
			flags=O_WRONLY|O_CREAT|O_TRUNC;
	}
	
	if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		char fname[UDMSTRSIZ]="";
		sprintf(fname,"%s%c%s",UDM_VAR_DIR,UDMSLASH,"dict");
#if (WIN32|WINNT)
		db->crcdict[0]=open(fname,flags);
#else
		db->crcdict[0]=open(fname,flags,0644);
#endif
		if(db->crcdict[0]<0){
			sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
		sprintf(fname,"%s%c%s",UDM_VAR_DIR,UDMSLASH,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}else
	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		int i;
		char fname[UDMSTRSIZ];
		for(i=1;i<MAXMULTI;i++){	
			sprintf(fname,"%s%cdict%02d",UDM_VAR_DIR,UDMSLASH,i);
#if (WIN32|WINNT)
			db->crcdict[i]=open(fname,flags);
#else
			db->crcdict[i]=open(fname,flags,0644);
#endif
			if(db->crcdict[i]<0){
				sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
				db->errcode=1;
				return(1);
			}
		}
		sprintf(fname,"%s%c%s",UDM_VAR_DIR,UDMSLASH,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}else{
		char fname[UDMSTRSIZ];
		sprintf(fname,"%s%c%s",UDM_VAR_DIR,UDMSLASH,"dict.txt");
		db->dict=fopen(fname,fmode);
		if(!db->dict){
			sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
		sprintf(fname,"%s%c%s",UDM_VAR_DIR,UDMSLASH,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}
	return(0);
}

void * UdmAllocDB(UDM_AGENT * Indexer, int mode){
DB * db;
	db=(DB*)malloc(sizeof(DB));
	db->dict=NULL;
	db->url=NULL;
	db->robots=NULL;
	db->stopwords=NULL;
	db->errcode=0;
	db->errmsg[0]=0;
	db->open_mode=mode;
	db->connected=0;
	return((void*)db);
}
char * UdmDBErrorMsg(void *db){
	return(((DB*)db)->errmsg);
}
int UdmDBErrorCode(void *db){
	return(((DB*)db)->errcode);
}


/*************  Words ****************************/
/*
typedef struct cstruct {
	int		url_id;
	udmcrc32_t	wrd_id;
	char		weight;
} UDM_CRCWORD;
*/

int UdmDeleteWordFromURL(UDM_AGENT* Indexer,int url_id){
	return(0);
}
int UdmStoreWords(UDM_AGENT * Indexer, int url_id, char * cat, int status){
	if (InitDB(Indexer)){
		UdmLog(Indexer, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
		exit(1);
	}
	if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		int fd,i,bytes;
		UDM_CRCWORD * cw;
		if(Indexer->nwords){
			fd=((DB*)(Indexer->db))->crcdict[0];
			bytes=sizeof(UDM_CRCWORD)*Indexer->nwords;
			cw=(UDM_CRCWORD*)malloc(bytes);
			for(i=0;i<Indexer->nwords;i++){
				if(Indexer->Word[i].count){
					cw[i].url_id=url_id;
					cw[i].weight=Indexer->Word[i].count,
					cw[i].wrd_id=UdmStrCRC32(Indexer->Word[i].word);
				}
			}
			if((bytes!=write(fd,(void *)cw,bytes))){
				sprintf(((DB*)(Indexer->db))->errmsg,"Can't write to dict file (%s)", strerror(errno));
				((DB*)(Indexer->db))->errcode=1;
				return(1);
			}
			free(cw);
		}
	}else
	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		int fd,i,j,len,bytes;
		UDM_CRCWORD * cw;
		if(Indexer->nwords){
			bytes=sizeof(UDM_CRCWORD)*Indexer->nwords;
			cw=(UDM_CRCWORD*)malloc(bytes);
			for(len=1;len<MAXMULTI;len++){
				j=0;
				for(i=0;i<Indexer->nwords;i++)
				if((Indexer->Word[i].count)&&
				    len==strlen(Indexer->Word[i].word)){
					cw[j].url_id=url_id;
					cw[j].weight=Indexer->Word[i].count,
					cw[j].wrd_id=UdmStrCRC32(Indexer->Word[i].word);
					j++;
				}
				fd=((DB*)(Indexer->db))->crcdict[len];
				bytes=sizeof(UDM_CRCWORD)*j;
				if((bytes!=write(fd,(void *)cw,bytes))){
					sprintf(((DB*)(Indexer->db))->errmsg,"Can't write to dict file (%s)", strerror(errno));
					((DB*)(Indexer->db))->errcode=1;
					return(1);
				}
			}
			free(cw);
		}
	}else{
		FILE *f;
		int i;
		f=((DB*)(Indexer->db))->dict;
		for(i=0;i<Indexer->nwords;i++){
			if(Indexer->Word[i].count){
				fprintf(f,"%d\t%d\t%s\n",url_id,
					Indexer->Word[i].count,
					Indexer->Word[i].word);
			}
		}
	}
	return(IND_OK);
}

/************** stopwords **********************/
int UdmLoadStopList(UDM_AGENT * Indexer){
	return(0);
}
UDM_STOPWORD * UdmIsStopWord(UDM_ENV * Conf,char *word){
	return(NULL);
}
void UdmFreeStopList(UDM_ENV * Env){

}
/************ URLs stuff *******************************/

static int n_urls=0;
static UDM_DOCUMENT* URLs=NULL;


UDM_DOCUMENT * UdmGetDocInfo(UDM_AGENT* Indexer,int expired_first){
int i;

UDM_DOCUMENT * Result;

	for(i=0;i<n_urls;i++){
		if(!URLs[i].indexed){
			URLs[i].indexed=1;
			Result=(UDM_DOCUMENT *)malloc(sizeof(UDM_DOCUMENT));
			Result->content_type=NULL;
			Result->title=NULL;
			Result->text=NULL;
			Result->last_index_time=0;
			Result->next_index_time=0;
			Result->last_mod_time=0;
			Result->keywords=NULL;
			Result->description=NULL;
			Result->url=strdup(URLs[i].url);
			Result->url_id=URLs[i].url_id;
			Result->size=0;
			Result->status=0;
			Result->hops=URLs[i].hops;
			Result->crc32=0;
			return(Result);
		}
	}
	return(NULL);
}
int UdmAddURL(UDM_AGENT* Indexer,char *url,int referrer,int hops, char * msg_id){
	if(!n_urls){
		URLs=(UDM_DOCUMENT *)malloc(sizeof(UDM_DOCUMENT));
	}else{
		URLs=(UDM_DOCUMENT *)realloc(URLs,sizeof(UDM_DOCUMENT)*(n_urls+1));
	}
	URLs[n_urls].url_id=n_urls+1;
	URLs[n_urls].indexed=0;
	URLs[n_urls].hops=hops;
	URLs[n_urls].referrer=referrer;
	URLs[n_urls].url=strdup(url);
	n_urls++;
	return(IND_OK);
}
int UdmDeleteUrl(UDM_AGENT* Indexer,int url_id){
	return(IND_OK);
}
int UdmUpdateUrl(UDM_AGENT* Indexer,int url_id,int status,int period){
	return(IND_OK);
}

#define XVAL(x)		(!x?"-":!x[0]?"-":x)
int UdmLongUpdateUrl(UDM_AGENT* Indexer,
	int url_id,int status,int is_changed,
	int size,int period,char * tag,int index,
	time_t last_mod_time,
	char *text_escaped,
	char *title_escaped,
	char *content_type,
	char *keywords_escaped,
	char *descript_escaped,
	udmcrc32_t crc32,
	char *lang,
	char *category
#ifdef NEWS_EXT
	,
	char *hd_date,
	char *hd_subj,
	char *hd_from,
	char *hd_group,
	char *hd_ref,
	char *msg_id
#endif
	){
	
	int i;
	FILE *f;

	if (InitDB(Indexer)){
	    UdmLog(Indexer, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
	    exit(1);
	}

	f=((DB*)(Indexer->db))->url;

	for(i=0;i<n_urls;i++)
	if(URLs[i].url_id==url_id){

		fprintf(f,"%d\t%s\t%d\t%d\t%s\t%s\t%u\t%s\t%li\t%s\t%s\t%s\t%s\n",
			url_id,
			XVAL(tag),status,size,
			XVAL(content_type),
			XVAL(lang),
			crc32,
			XVAL(URLs[i].url),
			last_mod_time,
			XVAL(text_escaped),
			XVAL(title_escaped),
			XVAL(keywords_escaped),
			XVAL(descript_escaped));
		break;
	}
	return(IND_OK);
}



/************* Clones ***************************/
int UdmFindOrigin(UDM_AGENT* Indexer, udmcrc32_t crc32, int size){
	return(0);
}

int UdmUpdateClone(UDM_AGENT* Indexer,int url_id,int status,int period,
	char *content_type,time_t last_mod_time,udmcrc32_t crc32){

	return(IND_OK);
}

UDM_DOCUMENT * UdmCloneList(UDM_AGENT * Indexer, udmcrc32_t crc32){
	return(NULL);
}
int UdmRegisterChild(UDM_AGENT *Indexer, int parent_id, int child_id){
	return 0;
}
int UdmFindMsgID(UDM_AGENT *Indexer, const char * msg_id){
	return 0;
}



/************* Clearing **************************/
int UdmDeleteAllFromDict(UDM_AGENT* Indexer){
	return(IND_OK);
}
int UdmDeleteAllFromUrl(UDM_AGENT* Indexer){
	return(IND_OK);
}
int UdmClearDB(UDM_AGENT* Indexer){
	return(IND_OK);
}



/************** Robots.txt stuff *****************/
int UdmLoadRobots(UDM_AGENT *Indexer){
	return(IND_OK);
}
int UdmDeleteRobotsFromHost(UDM_AGENT* Indexer,char *hostinfo){
	return(IND_OK);
}
int UdmAddRobotsToHost(UDM_AGENT* Indexer,char *hostinfo,char *s){
int nrobots;
	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
	if(!Indexer->Conf->Robots){
		nrobots=0;
		Indexer->Conf->Robots=(UDM_ROBOT *)malloc((2)*sizeof(UDM_ROBOT));
	}else{
		for(nrobots=0;Indexer->Conf->Robots[nrobots].hostinfo;nrobots++);
		Indexer->Conf->Robots=(UDM_ROBOT *)realloc(Indexer->Conf->Robots,(nrobots+2)*sizeof(UDM_ROBOT));
	}
	Indexer->Conf->Robots[nrobots].hostinfo=strdup(hostinfo);
	Indexer->Conf->Robots[nrobots].path=strdup(s);
	nrobots++;
	Indexer->Conf->Robots[nrobots].hostinfo=NULL;
	Indexer->Conf->Robots[nrobots].path=NULL;

	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
	return(IND_OK);
}



/**************** search stuff *******************/

/* Function to sort words in weight order */
static int cmpword(const void *s1,const void *s2){
int res;
	if(!(res=(((UDM_SEARCHWORD*)s2)->count)-(((UDM_SEARCHWORD*)s1)->count)))
		if(!(res=(((UDM_SEARCHWORD*)s2)->weight-(((UDM_SEARCHWORD*)s1)->weight))))
			if(!(res=(((UDM_SEARCHWORD*)s1)->url_id-(((UDM_SEARCHWORD*)s2)->url_id))));
	return(res);
}
/* Function to sort word list in URL_ID order */
static int cmpurlid(const void *s1,const void *s2){
	return(((UDM_SEARCHWORD*)s2)->url_id-((UDM_SEARCHWORD*)s1)->url_id);
}


UDM_DOCUMENT * UdmFind(UDM_AGENT * query, char * text){
char *w,*lasttok;
char *words[100];
udmcrc32_t cwords[100];
char str[UDMSTRSIZ];
char curword[100];
int weight;
int url_id;
int num_words=0;
int i,j;
UDM_SEARCHWORD * wrd=NULL;
int nwrd=0;
UDM_DOCUMENT * r;
int first;
int skip=0;
char * txt;

	if(InitDB(query))
		return(NULL);

	UdmLoadStopList(query);
	query->wordinfo[0]='\0';
	if(UdmDBErrorCode(query->db))return(NULL);
	txt=strdup(text);
	w=UdmGetWord(txt,&lasttok,query->Conf->local_charset);
	while(w){
		char ** ww;
		char *rw;

		ww=UdmNormalizeWord(query,w);
		rw=ww?*ww:w;

		if(UdmIsStopWord(query->Conf,rw)){
			sprintf(UDM_STREND(query->wordinfo)," %s :stopword",rw);
			w=UdmGetWord(NULL,&lasttok,query->Conf->local_charset);
			continue;
		}
		if(num_words<100){
			words[num_words]=strdup(rw);
			cwords[num_words]=UdmStrCRC32(rw);
			num_words++;
		}
		w=UdmGetWord(NULL,&lasttok,query->Conf->local_charset);
	}
	free(txt);

	if(query->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		UDM_CRCWORD cw[256];
		int bytes, wnum;
		while((bytes=read(((DB*)(query->db))->crcdict[0],&cw,sizeof(cw)))){
			wnum=bytes/sizeof(UDM_CRCWORD);
			for(j=0;j<wnum;j++){
				for(i=0;i<num_words;i++){
					if(cwords[i]==cw[j].wrd_id){
						if(!nwrd){
							nwrd=1;
							wrd=(UDM_SEARCHWORD*)malloc(nwrd*sizeof(UDM_SEARCHWORD));
						}else{
							nwrd++;
							wrd=(UDM_SEARCHWORD*)realloc(wrd,nwrd*sizeof(UDM_SEARCHWORD));
						}
						wrd[nwrd-1].url_id=cw[j].url_id;
						wrd[nwrd-1].count=1;
						wrd[nwrd-1].weight=cw[j].weight;
					}
				}
			}
		}
	}else
	if(query->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		UDM_CRCWORD cw[256];
		int i,j,bytes,len,fd,wnum;
		
		for(i=0;i<num_words;i++){
			len=strlen(words[i]);
			if(len>=MAXMULTI)len=MAXMULTI-1;
			fd=((DB*)(query->db))->crcdict[len];
			lseek(fd,0,SEEK_SET);
			while((bytes=read(fd,&cw,sizeof(cw)))){
				wnum=bytes/sizeof(UDM_CRCWORD);
				for(j=0;j<wnum;j++){
					if(cwords[i]==cw[j].wrd_id){
						if(!nwrd){
							nwrd=1;
							wrd=(UDM_SEARCHWORD*)malloc(nwrd*sizeof(UDM_SEARCHWORD));
						}else{
							nwrd++;
							wrd=(UDM_SEARCHWORD*)realloc(wrd,nwrd*sizeof(UDM_SEARCHWORD));
						}
						wrd[nwrd-1].url_id=cw[j].url_id;
						wrd[nwrd-1].count=1;
						wrd[nwrd-1].weight=cw[j].weight;
					}
				}
			}
		}
	}else{
		/* UDM_DBMODE_SINGLE */
		while(fgets(str,sizeof(str),((DB*)(query->db))->dict)){
			w=UdmGetToken(str,"\t",&lasttok);
			if(!w)continue;url_id=atoi(w);
			w=UdmGetToken(NULL,"\t",&lasttok);
			if(!w)continue;weight=atoi(w);
			w=UdmGetToken(NULL,"\t\n",&lasttok);
			if(!w)continue;strcpy(curword,w);
			for(i=0;i<num_words;i++){
				if(!strcmp(curword,words[i])){
					if(!nwrd){
						nwrd=1;
						wrd=(UDM_SEARCHWORD*)malloc(nwrd*sizeof(UDM_SEARCHWORD));
					}else{
						nwrd++;
						wrd=(UDM_SEARCHWORD*)realloc(wrd,nwrd*sizeof(UDM_SEARCHWORD));
					}
					wrd[nwrd-1].url_id=url_id;
					wrd[nwrd-1].count=1;
					wrd[nwrd-1].weight=weight;
				}
			}
		}
	}	
	/* Now let's sort in url_id order then group results */
	if(!nwrd)return(0);
	qsort((void*)wrd,nwrd,sizeof(UDM_SEARCHWORD),cmpurlid);

	j=0;
	for(i=1;i<nwrd;i++){
		if(wrd[j].url_id==wrd[i].url_id){
			wrd[j].weight+=wrd[i].weight;
			wrd[j].count++;
		}else{
			j++;
			wrd[j].url_id=wrd[i].url_id;
			wrd[j].count=wrd[i].count;
			wrd[j].weight=wrd[i].weight;
		}
	}
	nwrd=j+1;

	/* Sort groupped by url_id results in weight order */
	qsort((void*)wrd,nwrd,sizeof(UDM_SEARCHWORD),cmpword);

	if(query->search_mode==UDM_MODE_ALL){
		for(i=0;i<nwrd;i++){
			if(wrd[i].count<num_words){
				nwrd=i;
				break;
			}
			
		}
	}

	if(!nwrd){
		free(wrd);
		return(0);
	}
	
	r=(UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT)*(nwrd+1));
	for(i=0;i<nwrd;i++){
		r[i].url_id=wrd[i].url_id;
		r[i].rating=wrd[i].count;
		r[i].content_type=NULL;
		r[i].crc32=0;
		r[i].url=NULL;
		r[i].last_mod_time=0;
		r[i].text=NULL;
		r[i].title=NULL;
		r[i].keywords=NULL;
		r[i].description=NULL;
	}
	free(wrd);
	r[nwrd].url_id=0;

	while(fgets(str,sizeof(str),((DB*)(query->db))->url)){
		url_id=atoi(str);
		for(i=0;i<nwrd;i++){
			if(url_id==r[i].url_id){
				char * tag=NULL;
				int matched=1;
				w=UdmGetToken(str,"\t",&lasttok);
				for(j=0;w;j++){
					switch(j){
						case  1: tag=strdup(w);break;
						case  3: r[i].size=atoi(w);break;
						case  4: r[i].content_type=strdup(w);break;
						case  6: r[i].crc32=(udmcrc32_t)strtol(w, NULL, 10);break;
						case  7: r[i].url=strdup(w);break;
						case  8: r[i].last_mod_time=atol(w);break;
						case  9: r[i].text=strdup(w);break;
						case 10: r[i].title=strdup(w);break;
						case 11: r[i].keywords=strdup(w);break;
						case 12: r[i].description=strdup(w);break;
					}
					w=UdmGetToken(NULL,"\t",&lasttok);
				}

				/* Check "ul" subsection match  */
				if(query->Conf->ul)matched=(strstr(r[i].url,query->Conf->ul)!=NULL);
				/* Check "tag" subsection match */
				if((matched)&&(query->Conf->ttag)&&(tag))matched=(!strcmp(query->Conf->ttag,tag));

				if(!matched){
					/* Skip this URL */
					UDM_FREE(r[i].content_type);
					UDM_FREE(r[i].url);
					UDM_FREE(r[i].text);
					UDM_FREE(r[i].title);
					UDM_FREE(r[i].keywords);
					UDM_FREE(r[i].description);
					skip++;
					memmove(r+i,r+i+1,(nwrd-i)*sizeof(UDM_DOCUMENT));
				}
			}
		}
	}
	query->total_found=nwrd-skip;
	if(!query->total_found){
		free(r);
		return(NULL);
	}
	first=query->page_number*query->page_size;
	if(first>nwrd-skip)first=nwrd-skip;
	if(first>0)memmove(r,r+first,(nwrd-skip-first+1)*sizeof(UDM_DOCUMENT));
	if((nwrd-skip-first)>query->page_size)r[query->page_size].url_id=0;
	return(r);
}



/***************** Misc functions ***************/

int * UdmGetURLList(UDM_AGENT * query){
	return(NULL);
}

int UdmInsertSpell(UDM_AGENT *Indexer,char *flag,char *lang,char *word){
    return(0);
}

UDM_SPELL * UdmFindWordDB(UDM_AGENT * Indexer,char *word) {
    return(NULL);
}

int UdmInsertAffix(UDM_AGENT * Indexer,char flag,char *lang,char *mask,char *find,char *repl,char *type){
    return(0);
}

int UdmDBImportAffixes(UDM_AGENT * Indexer, int LCharset){
    return(0);
}

UDM_CATEGORY * UdmCatPath(UDM_AGENT * Indexer,char * category){
	return(NULL);
}

UDM_CATEGORY * UdmCatList(UDM_AGENT * Indexer,char * category){
	return(NULL);
}

__INDLIB__ int  UdmGetStatistics(UDM_AGENT * Indexer){
	return(IND_OK);
}
int UdmMarkForReindex(UDM_AGENT* Indexer){
	return(IND_OK);
}
__INDLIB__ int UdmAddTagLimit(UDM_ENV * Conf,char * tag){
	Conf->ttag=strdup(tag);
	return(0);
}
__INDLIB__ int UdmAddStatusLimit(UDM_ENV * Conf,int status){
	return(0);
}
__INDLIB__ int UdmAddURLLimit(UDM_ENV * Conf,char *URL){
	UDM_FREE(Conf->ul);
	Conf->ul=strdup(URL);
	return(0);
}
__INDLIB__ int UdmAddLangLimit(UDM_ENV * Conf,char *URL){
	return(0);
}

__INDLIB__ int UdmAddCatLimit(UDM_ENV * Conf,char * cat){
	return(0);
}

__INDLIB__ int UdmClearURLLimit(UDM_ENV * Conf){
	UDM_FREE(Conf->ul);
	UDM_FREE(Conf->ttag);
	return(0);
}
__INDLIB__ int UdmClearLimits(UDM_ENV * Conf){
	UDM_FREE(Conf->ul)
	return(0);
}
__INDLIB__ int UdmGetReferers(UDM_AGENT* Indexer){
	return(IND_OK);
}
__INDLIB__ int UdmAddTimeLimit(UDM_ENV * Conf,struct udm_stl_info_t * stl){
 return 0;
}


#endif
