#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#if (WIN32|WINNT)
#include <time.h>
#else
#include <sys/time.h>
#endif

#include "udm_utils.h"
#include "udm_common.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_robots.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_parseurl.h"
#include "udm_charset.h"
#include "udm_parser.h"
#include "udm_proto.h"
#include "udm_hrefs.h"
#include "udm_mutex.h"
#include "udm_crc32.h"
#include "udm_xmalloc.h"
#include "udm_http.h"
#include "udm_id3.h"
#include "udm_host.h"

#ifdef NEWS_EXT
#include "udm_parsedate.h"
#endif

#define UDM_THREADINFO(ThreadInfo,h,s,m)	if(ThreadInfo)ThreadInfo(h,s,m)

/**************** RFC 1522 ******************************************/
static int hex2int(char h){
	if((h>='0')&&(h<='9'))return(h-'0');
	if((h>='A')&&(h<='F'))return(h-'A'+10);
	if((h>='a')&&(h<='f'))return(h-'a'+10);
	return(0);
}
char * rfc1522_decode(char * dst, char *src){
char base64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
char *s,*d;

	s=src;d=dst;*dst=0;

	while(*s){
		char *e;

		if((e=strstr(s,"=?"))){
			char *schema;
			char *data;

			if(e>s){
				/* Copy last plain text buffer */
				strncpy(d,s,e-s);
				d+=(e-s);
				*d=0;
			}
			e+=2;
			if(!(schema=strchr(e,'?'))){
				/* Error */
				break;
			}
			schema++;
			data=schema+2;

			if(!(e=strstr(data,"?="))){
				/* This is an error */
				break;
			}
			switch(*schema){
				case 'Q':
				case 'q':
					while(data<e){
						char c;
						if(*data=='='){
							c=(char)hex2int(data[1])*16+hex2int(data[2]);
							data+=3;
						}else{
							c=data[0];
							data++;
						}
						/* Copy one char to dst */
						*d=c;d++;*d=0;
					}
					break;
				case 'B':
				case 'b':
					while(data<e){
						char *p;
						int x0,x1,x2,x3,res;

						p=strchr(base64,data[0]);x0=p?p-base64:0;
						p=strchr(base64,data[1]);x1=p?p-base64:0;
						p=strchr(base64,data[2]);x2=p?p-base64:0;
						p=strchr(base64,data[3]);x3=p?p-base64:0;

						res=x3+x2*64+x1*64*64+x0*64*64*64;

						p=(char*)(&res);

						if(p[2])*d=p[2];d++;*d=0;
						if(p[1])*d=p[1];d++;*d=0;
						if(p[0])*d=p[0];d++;*d=0;

						data+=4;
					}
					break;
				default:
					/* Error */
					schema="u";
					break;
			}
			if(*schema=='u'){
				/* Error */
				break;
			}
			s=e+2;
		}else{
			/* Copy plain text tail */
			strcpy(d,s);
			break;
		}
	}
	return(dst);
}

/********************  Words stuff *****************************/

#define RESORT_WORDS	256
#define WSIZE		1024

static int cmpwords(const void * v1, const void * v2){
	return(strcmp(((UDM_WORD*)v1)->word,((UDM_WORD*)v2)->word));
}
static int AddOneWord(UDM_AGENT *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
int i;
char *s;
UDM_STOPWORD * stop;
int l,r,c,res,wlen;

	wlen=strlen(word);
	if((wlen>Server->max_word_length)||(wlen<Server->min_word_length))
		return(0);
	s=strdup(word);
	if(strlen(s)>UDM_MAXWORDSIZE)s[UDM_MAXWORDSIZE]=0;
	if(checkstop){
		/* Stopwords checking and language guesser */
		if((stop=UdmIsStopWord(Indexer->Conf,s))){
			int i;
			UdmLog(Indexer,UDM_LOG_DEBUG,"stop: '%s'-'%s'", stop->word, stop->lang);
			/* Add language */
			if(!strcmp(stop->lang,""))
				Indexer->lang[0].count++;
			else
			for(i=1;i<UDM_LANGPERDOC;i++){
				if(!strcmp(Indexer->lang[i].lang,stop->lang)){
					Indexer->lang[i].count++;
					break;
				}
				/* if current lang is free */
				if(!strcmp(Indexer->lang[i].lang,"")){
					strcpy(Indexer->lang[i].lang,stop->lang);
					Indexer->lang[i].count++;
					break;
				}
			}
			free(s);
			return(0);
		}
	}
	/* Find current word in sorted part of word list */
	l=0;r=Indexer->swords-1;
	while(l<=r){
		c=(l+r)/2;
		res=strcmp(Indexer->Word[c].word,s);
		if(res==0){
			Indexer->Word[c].count|=where;
			free(s);
			return(0);
		}
		if(res<0)
			l=c+1;
		else
			r=c-1;
	}
	/* Find in unsorted part */
	for(i=Indexer->swords;i<Indexer->nwords;i++){
		if(!strcmp(Indexer->Word[i].word,s)){
			Indexer->Word[i].count|=where;
			free(s);return(0);
		}
	}
	/* Add new word */
	if(Indexer->nwords>=Indexer->mwords){
		if(Indexer->mwords){
			Indexer->mwords+=WSIZE;
			Indexer->Word=(UDM_WORD *)realloc(Indexer->Word,Indexer->mwords*sizeof(UDM_WORD));
		}else{
			Indexer->mwords=WSIZE;
			Indexer->Word=(UDM_WORD *)malloc(Indexer->mwords*sizeof(UDM_WORD));
		}
	}
	Indexer->Word[Indexer->nwords].word=s;
	Indexer->Word[Indexer->nwords].count=where;
	Indexer->nwords++;

	/* Sort unsorted part */
	if((Indexer->nwords-Indexer->swords)>RESORT_WORDS){
		qsort(Indexer->Word,Indexer->nwords,sizeof(UDM_WORD),cmpwords);
		Indexer->swords=Indexer->nwords;
	}
	return(0);
}

/* This function adds a normalized word form(s) into list using Ispell */
static int AddWord(UDM_AGENT *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
char 	** forms, ** saveforms;
int	have_digit=0;
int	have_alpha=0;

	if(Server->number_factor==0||Server->alnum_factor==0){
		char *s;
		s=word;
		while(*s){
			if(isdigit(*s))
				have_digit=1;
			else	
				have_alpha=1;
			if(have_digit&&have_alpha)break;
			s++;
		}
		if(have_digit){
			if(have_alpha){
				if(!Server->alnum_factor)return(0);
			}else{
				if(!Server->number_factor)return(0);
			}
		}
	}
	UdmTolower(word,Indexer->Conf->local_charset);
	if((saveforms=forms=UdmNormalizeWord(Indexer,word))){
		/* Add all NORMAL forms of the word */
		while(*forms){
			/* Add only if correct words are allowed */
			if(Server->correct_factor){
				AddOneWord(Indexer,Server,*forms,where,checkstop);
			}
			
			free(*forms);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMAL forms has not been found  */
		/*   then we will add the word itself  */
		/* Do it only when incorrect words are */
		/* allowed by configuration            */
		if(Server->incorrect_factor)
			AddOneWord(Indexer,Server,word,where,checkstop);
	}
	return(0);
}

static int FreeWords(UDM_AGENT* Indexer) {
	int i;
	for(i=0;i<Indexer->nwords;i++)
		free(Indexer->Word[i].word);
	Indexer->nwords=0;
	Indexer->swords=0;
	return(0);
}


/**************************** Built-in Parsers ***************************/

static int ParseText(UDM_AGENT * Indexer,UDM_SERVER * Server,
       char *content,int weight,int check_stopwords){
	char *s, *lt;
	if(weight&&content){
		s=UdmGetWord(content, &lt, Indexer->Conf->local_charset);
		while(s){
			AddWord(Indexer, Server, s, weight, check_stopwords);
			s=UdmGetWord(NULL, &lt, Indexer->Conf->local_charset);
		}
	}
	return(0);
}


/* HTML parser states */
#define HTML_TAG	1
#define HTML_TXT	2
#define HTML_COM	3

static int ParseHtml(
UDM_AGENT * Indexer,
UDM_SERVER * CurSrv,
char *content,
UDM_URL * curURL,
UDM_DOCUMENT * doc,
int index,int follow,
char *text,
char *keywords,
char *descript,
char *title){

	int inbody=0;
	int inscript=0;
	int instyle=0;
	int intitle=0;
	int hrefonly=0;
	int comment=0;
	unsigned int group=0;
	char *htok;
	char str[UDMSTRSIZ]="";
	char res[UDMSTRSIZ]="";

	strcpy(str, curURL->path);
	strcat(str, curURL->filename);
	hrefonly=(UDM_HREFONLY==UdmFilterType(Indexer->Conf,str,res));

	htok=content;

	while(*htok){
		char * href=NULL;
		char * tmp=NULL;
		char * hend;
		char * etmp;
		char * stmp;
		char * s;
		int have_space;
		int opening;
		UDM_TAG tag;
		int state=HTML_TXT;
		int len;


		if(!UDM_STRNCMP(htok,"<!--"))	state=HTML_COM;
		else	if(*htok=='<')		state=HTML_TAG;

		switch(state){
		case HTML_TAG: /* tag */

			href=NULL;
			for(hend=htok;(*hend!='>')&&(*hend);hend++);
			if(*hend=='>')hend++;
			tmp=(char*)malloc(hend-htok+1);
			strncpy(tmp,htok,hend-htok);tmp[hend-htok]=0;
			
			UdmParseTag(&tag,tmp);

			/* Convert to lower case */
			for(s=tag.tag;*s;*s=tolower(*s),s++);

			/* Detect whether opening or closing tag */
			if((opening=(tag.tag[0]!='/')))s=tag.tag;
			else				s=tag.tag+1;

			/* Let's find tag name in order of frequency */

			if(!strcmp(s,"a"))href=tag.href;/*117941*/
			else	if(!strcmp(s,"title"))	intitle=opening;/*6192*/
			else	
			if(!strcmp(s,"body")){/*5146*/
				if((inbody=opening)){
					int DCIndex;
#ifdef USE_CHARSET_GUESSER					
					DCIndex = UdmGuessCharset(content, UdmGetCharset(CurSrv->charset));
#else
					if(Indexer->charset){
						DCIndex=Indexer->charset;
					}else{
						DCIndex=UdmGetCharset(CurSrv->charset);
					}
					
#endif
					UdmRecode(content, DCIndex,Indexer->Conf->local_charset);
					UdmRecode(title,   DCIndex,Indexer->Conf->local_charset);
					UdmRecode(descript,DCIndex,Indexer->Conf->local_charset);
					UdmRecode(keywords,DCIndex,Indexer->Conf->local_charset);
				}
			}else
			if((!strcmp(tag.tag,"meta"))&&(tag.name)&&(tag.content)){
			
				if((!strcasecmp(tag.name,"Content-Type"))&&(!Indexer->charset)){
					char *p;
					if((p=strstr(tag.content,"charset="))){
						Indexer->charset = UdmGetCharset(p + 8);
					}
				}else
				if(!strcasecmp(tag.name,"refresh")){
					if((href=strstr(tag.content,"URL=")))
						href+=4;
				}else
				if(!hrefonly&&!strcasecmp(tag.name,"keywords")){
					strncpy(keywords,tag.content,UDM_MAXKEYWORDSIZE-1);
					keywords[UDM_MAXKEYWORDSIZE-1]=0;
					
				}else
				if(!hrefonly&&!strcasecmp(tag.name,"description")){
					strncpy(descript,tag.content,UDM_MAXDESCSIZE-1);
					descript[UDM_MAXDESCSIZE-1]=0;
					
				}else
				if(!strcasecmp(tag.name,"robots")&&CurSrv->userobots&&tag.content){
					char * lt;
					char * s;
					s=UdmGetWord(tag.content,&lt,Indexer->Conf->local_charset);
					while(s){
						if(!strcasecmp(s,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(s,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(s,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(s,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(s,"INDEX"))
							index=1;
						else
						if(!strcasecmp(s,"FOLLOW"))
							follow=1;
						s=UdmGetWord(NULL,&lt,Indexer->Conf->local_charset);
					}
				}else
				if((!strcasecmp(tag.name,"language"))&&(tag.content)){
					strncpy(Indexer->lang[1].lang,tag.content,2);
					Indexer->lang[1].lang[2]=0;
					Indexer->lang[1].count=1000000; /* This should be enough */
				}
			}
			else	if(!strcmp(s,"img"))	href=tag.src;/*2786*/
			else	if(!strcmp(s,"link"))	href=tag.href;/*2241*/
			else	if(!strcmp(s,"frame"))	href=tag.src;
			else	if(!strcmp(s,"script"))	inscript=opening;
			else	if(!strcmp(s,"style"))	instyle=opening;
			else	if(!strcmp(s,"area"))	href=tag.href;
			else	if(!strncasecmp(s,"group_", 6)){
					if(opening)
						if((group = atoi(s+6))>0)
							group=1<<group;
						else
							group = 0;
					else
						group = 0;
				}
			else	if((!strcmp(s,"base"))&&(tag.href)){
				/* <BASE HREF="xxx"> stuff            */
				/* Check that URL is properly formed  */
				/* baseURL is just temporary variable */
				/* If parsing is fail we'll use old   */
				/* base href, passed via curURL       */
				/* The last is surely correct here    */
				
				/* Note that we will not check BASE     */
				/* if delete_no_server is unset         */
				/* This is  actually dirty hack. We     */
				/* must check that hostname is the same */
				
				
				if((CurSrv->delete_no_server)||(CurSrv->outside)){
					UDM_URL baseURL;
					int res;
					
					if(!(res=UdmParseURL(&baseURL,tag.href))){
						UdmParseURL(curURL,tag.href);
					}else{
						switch(res){
						case UDM_PARSEURL_LONG:
							UdmLog(Indexer,UDM_LOG_ERROR,"BASE HREF too long: '%s'",tag.href);
							break;
						case UDM_PARSEURL_BAD:
						default:
							UdmLog(Indexer,UDM_LOG_ERROR,"Error in BASE HREF URL: '%s'",tag.href);
						}
					}
				}
			}

			if((href)&&(follow)&&(CurSrv->gfollow)){
				UDM_URL newURL;
				int res;
				if(doc->hops>=CurSrv->maxhops){
					UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : too many hops: %d",href,doc->hops);
				}else
				if(!(res=UdmParseURL(&newURL,href))){
					char * newschema;
					int Method;
					char reason[UDMSTRSIZ]="";
					if(newURL.schema[0])newschema=newURL.schema;
					else	newschema=curURL->schema;
					if(!strcasecmp(newschema,"file")||!strcasecmp(newschema,"htdb")){
						sprintf(str,"%s:%s%s",newschema,newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}else{
						sprintf(str,"%s://%s%s%s",
						newURL.schema[0]?newURL.schema:curURL->schema,
						newURL.hostinfo[0]?newURL.hostinfo:curURL->hostinfo,
						newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}
					if((newschema=strchr(str,':')))
						UdmRemove2Dot(newschema+1);
					
					if(!UDM_STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;

					Method=UdmFilterType(Indexer->Conf,str,reason);
					UdmLog(Indexer,UDM_LOG_DEBUG,"\"%s\" : '%s'",href,reason);

					if((Method==UDM_DISALLOW)){
					}else
					if((UdmFindRobots(Indexer->Conf,str))&&(CurSrv->userobots)){
						UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : robots",href);
						UdmLog(Indexer,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if(!UdmFindServer(Indexer->Conf,str)){
						UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : no Server",href);
					}else{
						int add=1;
						/* compare hostinfo in some cases */
						if((!CurSrv->delete_no_server)&&(!CurSrv->outside)){
							if(newURL.hostinfo[0])
								add=!strcmp(curURL->hostinfo,newURL.hostinfo);
						}
						if(add){
						
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
							
							/* Add URL itself */
							if(UdmAddHref(Indexer->Conf,str,doc->url_id,doc->hops+1,0))
								Indexer->Conf->have_targets=1;
							
							/* Add robots.txt for HTTP schema */
							/* When FollowOutside or DeleteNoServer no */
							if((!strcasecmp(newURL.schema,"http"))&&(CurSrv->userobots)&&
							((CurSrv->outside)||(!CurSrv->delete_no_server))){
								char str1[UDMSTRSIZ]="";
								sprintf(str1,"%s://%s/%s",newURL.schema,newURL.hostinfo,"robots.txt");
								if(UdmAddHref(Indexer->Conf,str1,0,0,0))
									Indexer->Conf->have_targets=1;
							}
							
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
						}
					}
				}else{
					switch(res){
					case UDM_PARSEURL_LONG:
						UdmLog(Indexer,UDM_LOG_DEBUG,"URL too long: '%s'",href);
						break;
					case UDM_PARSEURL_BAD:
					default:
						UdmLog(Indexer,UDM_LOG_DEBUG,"Error in URL: '%s'",href);
					}
				}
			}
			UdmFreeTag(&tag);
			free(tmp);
			htok=hend;
			break;
 
		case HTML_COM: /* comment */

			if(!UDM_STRNCASECMP(htok,"<!--UdmComment-->"))
				comment=1;
			else
			if(!UDM_STRNCASECMP(htok,"<!--/UdmComment-->"))
				comment=0;

			while(*htok){
				if(!UDM_STRNCMP(htok,"-->")){
					htok+=3;
					break;
				}
				htok++;
			}
			break;

		case HTML_TXT: /* text */

			if(inscript){
				/* Special case when script  */
				/* body is not commented:    */
				/* <script> x="<"; </script> */
				/* We should find </script>  */
				/* in this case:             */
				for(hend=htok;*hend;hend++){
					if(!UDM_STRNCASECMP(hend,"</script>"))
						break;
				}
			}else{
				for(hend=htok;(*hend)&&(*hend!='<');hend++);
			}
			tmp=(char*)malloc(2*(hend-htok)+6);
			etmp=tmp;*etmp=0;have_space=0;
			for(stmp=htok;stmp<hend;stmp++){
				char sch;
				switch(*stmp){
				case ' ' : case '\n':
				case '\r': case '\t':
					have_space++; break;

				case '&': /* parse specials */
					if(have_space){
						*etmp=' ';etmp++;
						have_space=0;
					}
					stmp++;
					if(*stmp=='#'){ /* &#234; */
						char *code;
						code=++stmp;
						while((stmp<hend)&&(isdigit(*(stmp+1))))
							stmp++;
						if(*(stmp+1)==';')stmp++;
						*etmp=(char)(atoi(code));
						etmp++;*etmp=0;
					}else
					if((sch=UdmSgmlToChar(stmp,Indexer->Conf->local_charset))){
						if(sch==' ')have_space++;
						*etmp=sch;etmp++;*etmp=0;
						while((stmp<hend)&&(isalnum(*(stmp+1))))
							stmp++;
						if(*(stmp+1)==';')stmp++;
					}else{
						*etmp='?';etmp++;*etmp=0;
					}
					break;
				default:
					if(have_space){
						*etmp=' ';etmp++;
						have_space=0;
					}
					*etmp=*stmp;
					etmp++;*etmp=0;
					break;
				}
			}
			if((!comment&&inbody&&!inscript&&!instyle)&&((len=strlen(text))<UDM_MAXTEXTSIZE-2)){
				len=UDM_MAXTEXTSIZE-2-len;
				if(*text)strcat(text," ");
				strncat(text,tmp,len);
				text[UDM_MAXTEXTSIZE-1]=0;
			}
			if((!comment)&&(intitle)&&((len=strlen(title))<UDM_MAXTITLESIZE-2)){
				len=UDM_MAXTITLESIZE-2-len;
				if(*title)strcat(title," ");
				strncat(title,tmp,len);
				title[UDM_MAXTITLESIZE-1]=0;
			}
			if(!intitle&&!comment&&(!hrefonly||group)&&index&&!inscript&&!instyle){
				ParseText(Indexer,CurSrv,tmp,group?group:CurSrv->bodyweight*inbody,1);
			}
			free(tmp);
			htok=hend;
			break;
		}
	}
	if(index&&CurSrv->keywordweight){
		strcpy(str,title);
		ParseText(Indexer,CurSrv,str,CurSrv->titleweight,1);
	}
	if(index&&CurSrv->keywordweight){
		strcpy(str,keywords);
		ParseText(Indexer,CurSrv,str,CurSrv->keywordweight,1);
	}
	if(index&&CurSrv->descweight){
		strcpy(str,descript);
		ParseText(Indexer,CurSrv,str,CurSrv->descweight,1);
	}
	return(0);
}

/****************************************************************/
static int cmplang(const void *s1,const void *s2){
	return(((UDM_LANG*)s2)->count-((UDM_LANG*)s1)->count);
}


static void FreeDoc(UDM_DOCUMENT *Result){
	if(!Result)return;
	UDM_FREE(Result->url);
	UDM_FREE(Result->content_type);
	UDM_FREE(Result->title);
	UDM_FREE(Result->text);
	UDM_FREE(Result->keywords);
	UDM_FREE(Result->description);
	free(Result);
}


/*********************** 'UrlFile' stuff (for -f option) *******************/

__INDLIB__ int UdmURLFile(UDM_AGENT *Indexer, int action){
FILE *url_file;
char str[UDMSTRSIZ]="";
char str1[UDMSTRSIZ]="";
int result,res;
UDM_URL myurl;

	/* Read lines and clear/insert/check URLs                     */
	/* We've already tested in main.c to make sure it can be read */
	/* FIXME !!! Checking should be done here surely              */

	if(!strcmp(Indexer->Conf->url_file_name,"-"))
		url_file=stdin;
	else
		url_file=fopen(Indexer->Conf->url_file_name,"r");

	while(fgets(str1,sizeof(str1),url_file)){
		char *end;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		switch(action){
		case UDM_URL_FILE_REINDEX:
			UdmAddURLLimit(Indexer->Conf,str);
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
			result=UdmMarkForReindex(Indexer);
			if(result!=IND_OK)return(result);
			UdmClearURLLimit(Indexer->Conf);
			break;
		case UDM_URL_FILE_CLEAR:
			UdmAddURLLimit(Indexer->Conf,str);
			result=UdmClearDB(Indexer);
			if(result!=IND_OK)return(IND_ERROR);
			UdmClearURLLimit(Indexer->Conf);
			break;
		case UDM_URL_FILE_INSERT:
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			if(UdmAddHref(Indexer->Conf,str,0,0,0))
				Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
			
			break;
		case UDM_URL_FILE_PARSE:
			res=UdmParseURL(&myurl,str);
			if((!res)&&(!myurl.schema[0]))
				res=UDM_PARSEURL_BAD;
			if(res){
				switch(res){
				case UDM_PARSEURL_LONG:
					UdmLog(Indexer,UDM_LOG_ERROR,"URL too long: '%s'",str);
					break;
				case UDM_PARSEURL_BAD:
				default:
					UdmLog(Indexer,UDM_LOG_ERROR,"Error in URL: '%s'",str);
				}
				return(IND_ERROR);
			}
			break;
		}
		str[0]=0;
	}
	if(url_file!=stdin)
		fclose(url_file);
	return(IND_OK);
}


/******* Main indexer functions StartUp & IndexNextURL ********/

/* Init section for indexing */

static int StartUp(UDM_AGENT * Indexer, int index_flags){
	Indexer->Conf->have_targets=1;
	if(index_flags&UDM_FLAG_MARK){
		if(Indexer->Conf->url_file_name)
			UdmURLFile(Indexer,UDM_URL_FILE_REINDEX);

		UdmMarkForReindex(Indexer);
		if(UdmDBErrorCode(Indexer->db)){
			return(1);
		}
	}
	if((Indexer->Conf->url_file_name) && (index_flags&UDM_FLAG_INSERT)) {
		UdmURLFile(Indexer,UDM_URL_FILE_INSERT);
		if(UdmDBErrorCode(Indexer->db)){
			return(1);
		}
	}
	UdmLoadRobots(Indexer);
	if(UdmDBErrorCode(Indexer->db)){
		return(1);
	}
	UdmLoadStopList(Indexer);
	if(UdmDBErrorCode(Indexer->db)){
		return(1);
	}
	if(index_flags&UDM_FLAG_SKIP_LOCKING)
		Indexer->Conf->DBUseLock=0;
	else
		Indexer->Conf->DBUseLock=1;
	return(0);
}


__INDLIB__ int UdmIndexNextURL(UDM_AGENT *Indexer,int index_flags){
char request[UDMSTRSIZ]="";
char reason[UDMSTRSIZ]="";
udmcrc32_t crc32;

char *lt,*tok;
char *content=NULL,*header,*content_type,*location,*statusline;
char *hcopy=NULL;

int size,status=0,realsize,Method=0;
int origin,follow,index,changed;
int reindex;
int found_in_mirror = 0;
int result=IND_UNKNOWN;

UDM_URL		curURL;
UDM_URL		realURL;
UDM_DOCUMENT	*Doc=NULL;
UDM_SERVER	*CurSrv=0;
UDM_ALIAS	*Alias;

char text[2*UDM_MAXTEXTSIZE]="";
char text_escaped[2*UDM_MAXTEXTSIZE+1]="";
char keywords[UDM_MAXKEYWORDSIZE]="";
char keywords_escaped[2*UDM_MAXKEYWORDSIZE+1]="";
char descript[UDM_MAXDESCSIZE]="";
char descript_escaped[2*UDM_MAXDESCSIZE+1]="";
time_t last_mod_time=0;
char buf[UDM_MAXTIMESTRLEN]="";
char title[UDM_MAXTITLESIZE]="";	
char title_escaped[UDM_MAXTITLESIZE*2+1];
char subj[UDM_MAXTITLESIZE]="";
char from[UDM_MAXKEYWORDSIZE]="";
char content_type_escaped[256]="";

#ifdef NEWS_EXT
/* I need to retain newsgroup-info, so I introduce new variables */
/* I also stick with these fixed-length strings, although I dislike them.*/
char HeaderDate[UDM_MAXDATESIZE] = "";
char HeaderDateEsc[2*UDM_MAXDATESIZE+1] = "";
char HeaderSubject[UDM_MAXSUBJSIZE] = "";
char HeaderSubjectEsc[2*UDM_MAXSUBJSIZE+1] = "";
char HeaderFrom[UDM_MAXFROMSIZE] = "";
char HeaderFromEsc[2*UDM_MAXFROMSIZE+1] = "";
char HeaderGroup[UDM_MAXGROUPSIZE] = "";
char HeaderGroupEsc[2*UDM_MAXGROUPSIZE+1] = "";
char HeaderRefs[UDM_MAXREFSIZE] = "";
char HeaderRefsEsc[2*UDM_MAXREFSIZE+1] = "";
char MessageIdEsc[2*UDM_MAXFROMSIZE+1]="";
char *parent=NULL, *SQLDate;
#endif

#ifdef USE_PARSER
char *mime;
int mimeno;
#endif
int is_mp3=0, mp3_size=0;
	/* Do init stuff if required */
	if(index_flags&UDM_FLAG_INIT){
		if(StartUp(Indexer, index_flags))
			return(IND_ERROR);
		else
			return(IND_OK);
	}else{
		int j;
		/* Clear language statistics */
		for(j=0;j<UDM_LANGPERDOC;j++){
			Indexer->lang[j].lang[0]=0;
			Indexer->lang[j].count=0;
		}
	}

	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
	/* Store URLs from cache into database */
	if(UdmStoreHrefs(Indexer))
		Indexer->Conf->have_targets=1;
	if(UdmDBErrorCode(Indexer->db)){
		result=IND_ERROR;
	}else
	if(!Indexer->Conf->have_targets){
		result=IND_NO_TARGET;
	}else{
		/* Get Next URL to be indexed from the database */
		UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Selecting","");
		if(!(Doc=UdmGetDocInfo(Indexer,index_flags))){
			if(UdmDBErrorCode(Indexer->db)){
				result=IND_ERROR;
			}else{
				Indexer->Conf->have_targets=0;
				result=IND_NO_TARGET;
			}
		}else{
			/* Add current URL in memory cache to avoid */
			/* possible double INSERT INTO url          */
			UdmAddHref(Indexer->Conf,Doc->url,0,Doc->hops,1);
			Indexer->Conf->have_targets=1;
		}
	}
	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
	/* In GUI version we may send "terminate" signal       */
	/* This line checks whether it already has happened    */
	/* It allows terminate indexing faster                 */
	/* We include this line in several places of this func */
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;
	if(result)return(result);

	/* Alloc buffer for document */
	if(!Indexer->buf){
		Indexer->buf=(char*)malloc(Indexer->Conf->max_doc_size);
	}

	UdmLog(Indexer,UDM_LOG_INFO,"%s",Doc->url);
	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Found",Doc->url);
	reindex=index_flags&UDM_FLAG_REINDEX;

#ifdef HAVE_SETPROCTITLE

	/* To see the URL being indexed   */
	/* in "ps" output on FreeBSD      */
	/* Do it if single thread version */

	if(!(Indexer->handle))
		setproctitle("%s",Doc->url);
#endif

	/* Check that URL has valid syntax */
	if(UdmParseURL(&curURL,Doc->url)){
		result=UdmDeleteUrl(Indexer,Doc->url_id);
		UdmLog(Indexer,UDM_LOG_WARN,"Invalid URL: %s ... deleting",Doc->url);
		FreeDoc(Doc);
		return(result);
	}

	/* Find correspondent Server record from indexer.conf */
	if(!(CurSrv=UdmFindServer(Indexer->Conf,Doc->url))){
		if(!strcmp(curURL.filename,"robots.txt")){
			if(IND_OK==(result=UdmDeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=UdmLoadRobots(Indexer);
		}else{
			UdmLog(Indexer,UDM_LOG_WARN,"No 'Server' command for url... deleted.");
			result=IND_OK;
		}
		if(result==IND_OK)result=UdmDeleteUrl(Indexer,Doc->url_id);
		FreeDoc(Doc);
		return(result);
	}

	Indexer->read_timeout=CurSrv->read_timeout;
	Indexer->maxsize=Indexer->Conf->max_doc_size;

	/* Check that hops is less than MaxHops */
	if(Doc->hops>CurSrv->maxhops){
		result=UdmDeleteUrl(Indexer,Doc->url_id);
		UdmLog(Indexer,UDM_LOG_WARN,"Too many hops (%d)... deleted.",Doc->hops);
		FreeDoc(Doc);
		return(result);
	}

	/* Check Allow/Disallow/CheckOnly stuff */

	Method=UdmFilterType(Indexer->Conf,Doc->url,reason);
	UdmLog(Indexer,UDM_LOG_DEBUG,"'%s'",reason);

	if(Method==UDM_DISALLOW){
		result=UdmDeleteUrl(Indexer,Doc->url_id);
		if((result==IND_OK)&&(!strcmp(curURL.filename,"robots.txt"))){
			if(IND_OK==(result=UdmDeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=UdmLoadRobots(Indexer);
		}
		FreeDoc(Doc);
		return(result);
	}

	UdmParseURL(&realURL,Doc->url);
	/* Check whether URL is disallowed by robots.txt */
	if((CurSrv->userobots)&&(UdmFindRobots(Indexer->Conf,Doc->url))){
		UdmLog(Indexer,UDM_LOG_WARN,"Deleting URL: robots.txt");
		result=UdmDeleteUrl(Indexer,Doc->url_id);
	}else
	/* Check for too many errors on this server */
	if(UdmGetServerErr(Indexer->Conf,realURL.hostname)>CurSrv->max_net_errors){
		UdmLog(Indexer,UDM_LOG_WARN,"Too many network errors for this server, skipped");
		result=UdmUpdateUrl(Indexer,Doc->url_id,504,CurSrv->net_error_delay_time);
	}
	if(result){
		FreeDoc(Doc);
		return(result);
	}

	/* Find alias */
	if((Alias=UdmFindAlias(Indexer->Conf,Doc->url))){
		char rurl[UDMSTRSIZ];
		sprintf(rurl,"%s%s",Alias->replace,Doc->url+strlen(Alias->find));
		if(UdmParseURL(&realURL,rurl)){
			UdmLog(Indexer,UDM_LOG_ERROR,"Error in aliased URL: '%s'",rurl);
			FreeDoc(Doc);
			return(IND_OK);
		}else{
			UdmLog(Indexer,UDM_LOG_EXTRA,"Alias: '%s'",rurl);
		}
	}


	/* Compose HTTP/1.0 request header */
	sprintf(request,"%s %s%s HTTP/1.0\r\n",
		(Method==UDM_HEAD)?"HEAD":"GET",
		CurSrv->proxy?Doc->url:realURL.path,
		CurSrv->proxy?"":realURL.filename);

	/* Add If-Modified-Since header */
	if((Doc->status)&&(Doc->last_mod_time)&&(!reindex)){
		UdmTime_t2HttpStr(Doc->last_mod_time, buf);
		sprintf(UDM_STREND(request),"If-Modified-Since: %s\r\n",
			buf);
	}

	/* HTTP and FTP specific stuff */
	if((!strcasecmp(realURL.schema,"http"))||(!strcasecmp(realURL.schema,"ftp"))){

		/* User agent */
		sprintf(UDM_STREND(request),"User-Agent: %s\r\n",Indexer->Conf->user_agent);

		/* If LocalCharset specified */
		if( Indexer->Conf->local_charset )
			sprintf(UDM_STREND(request),"Accept-charset: %s\r\n",
				(char*)UdmCharsetStr(Indexer->Conf->local_charset));

		/* Host Name for virtual hosts */
		sprintf(UDM_STREND(request),"Host: %s\r\n",realURL.hostname);

		/* Auth if required */
		if(CurSrv->basic_auth)
			sprintf(UDM_STREND(request),"Authorization: Basic %s\r\n",
				CurSrv->basic_auth);
		
		/* ProxyAuth if required */
		if(CurSrv->proxy_basic_auth)
			sprintf(UDM_STREND(request),"Proxy-Authorization: Basic %s\r\n",
				CurSrv->proxy_basic_auth);
	}

	/* Add user defined headers */
	strcat(request,Indexer->Conf->extra_headers);

	/* Empty line is the end of HTTP header */
	strcat(request,"\r\n");

	size=UDM_NET_UNKNOWN;

	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Getting",Doc->url);

	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;

	/* If mirroring is enabled */
	if (CurSrv->use_mirror >= 0) {
		char errstr[UDMSTRSIZ]="";
		/* on u_m==0 it returned by mtime from mirrorget */
		/* but we knew that it should be put in mirror  */

		size = UdmMirrorGET(Indexer,CurSrv->mirror_root,CurSrv->mirror_headers,
				realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				Indexer->buf, Indexer->Conf->max_doc_size,
				CurSrv->use_mirror,errstr);
		if(size>0){
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s has been taken from mirror",Doc->url);
			found_in_mirror=1;
		}else{
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s",errstr);
		}
	}

	/* Get it from the source  */
	/* if no mirror copy found */

	if(!found_in_mirror){
#ifdef USE_HTDB
		if(!strcasecmp(realURL.schema,"htdb")){
			size = UdmHTDBGet(Indexer,realURL.path,realURL.filename,CurSrv->htdb_list,CurSrv->htdb_doc);
			if(UdmDBErrorCode(Indexer->db))result=IND_ERROR;
		}
#endif
#ifdef USE_FILE
		if(!strcasecmp(realURL.schema,"file")){
			size = UdmFILEGet(Indexer,request);
		}
#endif
#ifdef USE_NEWS
		if((!strcasecmp(realURL.schema,"news"))){
			size = UdmNEWSGet(Indexer,request,realURL.hostname,realURL.port?realURL.port:119);
		}else
		if((!strcasecmp(realURL.schema,"nntp"))){
			size = UdmNNTPGet(Indexer,request,realURL.hostname,realURL.port?realURL.port:119);
		}
#endif
#ifdef USE_HTTP
		if((!strcasecmp(realURL.schema,"http")) ||
		    ((!strcasecmp(realURL.schema,"ftp"))&&(CurSrv->proxy))){
			int get_doc=1;
#ifdef USE_MP3
			if (CurSrv->check_mp3_tag && curURL.filename[0] !='\0'){
				int is_del=0, i;
				char range[16];
				UDM_HTTP_HDR *http_hdr=NULL;

				sprintf(range, "0-256");
				for( i=0; i<2; i++){
					http_hdr_free(http_hdr);
					http_hdr=NULL;

					Indexer->maxsize=10240;
					if ((size = UdmHTTPGetRange(Indexer,request,
					    CurSrv->proxy?CurSrv->proxy:realURL.hostname,
					    CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80),
					    range)) ==-1){
						break;
					}
				
					http_hdr = http_parse_header(Indexer->buf);
					if (http_hdr && http_hdr->status==206){
					        if ((content = get_id3_tag(http_hdr->body_ptr,
									size-http_hdr->hdr_size))==NULL){
							if (i==0)
								sprintf(range, "-128");
							else{
							/* It's mp3 file but no ID3 tag here */
								is_mp3=1; get_doc=0; is_del=0;
								break;
							}
						}else if (content == (char*)-1){
							if (i==1){
							/* It's mp3 file but no ID3 tag here */
								is_mp3=1; get_doc=0; is_del=0;
								break;
							}else{
							/* Not mp3 file */
								is_mp3=0; get_doc=1; is_del=1;
								break;
							}
						}else{
							/* It's mp3 file and id3 tag here */
							is_mp3=1; get_doc=0; is_del=0;
						}
					}else if (http_hdr && http_hdr->status==200){
						/* Fix me: Range don't work why? */
						is_mp3=0; get_doc=1; is_del=0;
						break;
					}else{
						/* Some http error */
						is_mp3=0; get_doc=0; is_del=0;
						break;
					}
				}

				if(is_mp3){
					char *fl_str;
					mp3_size=http_hdr->content_range;
					memcpy(http_hdr->body_ptr-2, "Content-type: text/html\r\n\r\n", 27);
					http_hdr->body_ptr += 25;
					fl_str = UdmXmalloc(strlen(curURL.filename)+1);
					if(content && content != (char*)-1){
						udm_snprintf(http_hdr->body_ptr, Indexer->Conf->max_doc_size-http_hdr->hdr_size,
					    		"%s<group_4>%s</group_4>", content,UdmUnescapeCGIQuery( fl_str ,curURL.filename));
						UDM_FREE(content);
					}else{
						udm_snprintf(http_hdr->body_ptr, Indexer->Conf->max_doc_size-http_hdr->hdr_size,
					    		"<group_4>%s</group_4>", UdmUnescapeCGIQuery( fl_str ,curURL.filename));
					}
					size = strlen(Indexer->buf);
					UDM_FREE(fl_str);
				}

				http_hdr_free(http_hdr);
				http_hdr=NULL;

				if (CurSrv->check_only_mp3_tag && !is_mp3 && is_del){
					result=UdmDeleteUrl(Indexer,Doc->url_id);
					UdmLog(Indexer,UDM_LOG_WARN,"No ID3 tag detected in %s... deleted.",
					realURL.specific);
					FreeDoc(Doc);
					return(result);
				}
			}	
#endif
			if (get_doc){
				size = UdmHTTPGet(Indexer,request,
					CurSrv->proxy?CurSrv->proxy:realURL.hostname,
					CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80));
			}
		}
#endif
#ifdef USE_FTP
		if ((!strcasecmp(curURL.schema,"ftp"))&&(!CurSrv->proxy)){
		    size = UdmFTPGet(Indexer,realURL.hostname,realURL.port?realURL.port:21,
			curURL.path, curURL.filename[0]=='\0'?NULL:curURL.filename, 
			CurSrv->user,CurSrv->passwd,Doc->last_mod_time,
			(Method==UDM_HEAD?1:0));
		}
#endif
	}

	if(!result) /* Extra check for HTDB possible bad result */
	switch(size){
	case UDM_NET_UNKNOWN:
		UdmLog(Indexer,UDM_LOG_WARN,"Protocol not supported");
		status=UDM_HTTP_STATUS_NOT_SUPPORTED;
		result=IND_OK;
		break;
	case UDM_NET_TIMEOUT:
		UdmLog(Indexer,UDM_LOG_WARN,"Download timeout");
		status=UDM_HTTP_STATUS_TIMEOUT;
		result=IND_OK;
		break;
	case UDM_NET_CANT_CONNECT:
		UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",
		CurSrv->proxy?CurSrv->proxy:curURL.hostname,
		CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
		status=UDM_HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	case UDM_NET_CANT_RESOLVE:
		UdmLog(Indexer,UDM_LOG_WARN,"Unknown host %s",
			CurSrv->proxy?CurSrv->proxy:curURL.hostname);
		status=UDM_HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	default:
		if(size<0){	/* No connection */
			UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",
				CurSrv->proxy?CurSrv->proxy:curURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
			status=UDM_HTTP_STATUS_UNAVAIL;
			result=IND_OK;
		}else{
			/* Document has been successfully loaded */
			/* Cut HTTP response header first        */
			if((content=strstr(Indexer->buf,"\r\n\r\n"))){
				*content = 0;
				content +=4;
			} else if ((content=strstr(Indexer->buf,"\n\n"))){
				*content = 0;
				content +=2;
			}
			if(!content){
				UdmLog(Indexer,UDM_LOG_ERROR,"Illegal HTTP headers in response");
				status=UDM_HTTP_STATUS_UNAVAIL;
				result=IND_OK;
			}
		}
	}

	/* Check again whether "terminate" was pressed */
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;

	/* Exit if there was an error while downloading */
	if(result){
		UdmAddServerErr(Indexer->Conf,realURL.hostname);
		if(result!=IND_ERROR)
			result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->net_error_delay_time);
		FreeDoc(Doc);
		return(result);
	}

	/* Let's start parsing */
	Indexer->buf[size]=0;header=Indexer->buf;
	realsize=size;
	status=changed=1;origin=0;
	content_type=location=statusline=NULL;
	Indexer->charset=UDM_CHARSET_USASCII;
	crc32=0;text[0]=0;title[0]=0;keywords[0]=0;
	descript[0]=0;subj[0]=0;last_mod_time=0;from[0]=0;
	last_mod_time=Doc->last_mod_time;

	index	=CurSrv->gindex;
	follow	=CurSrv->gfollow;

	size-=(content-Indexer->buf);     /* Could be modified by Content-Length   */
	realsize-=(content-Indexer->buf); /* Will be safe for crc32, Parsers etc   */

	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Parsing",Doc->url);
	/* Now lets parse response header lines */
	hcopy = strdup(header);
	tok=UdmGetToken(header,"\r\n",&lt);
	while(tok){
		if(!UDM_STRNCASECMP(tok,"HTTP/")){
			status=atoi(tok+8);
			statusline=tok;
		}else
		if ((Indexer->Conf->force1251) && (!UDM_STRNCASECMP(tok,"Server: "))){
			char * sname;
			sname=UdmTrim(tok+7," ");
			if (!UDM_STRNCASECMP(sname,"Microsoft")||
				!UDM_STRNCASECMP(sname,"IIS"))
					Indexer->charset=UDM_CHARSET_CP1251;
		}else
		if(!UDM_STRNCASECMP(tok,"Content-Type: ")){
			char *p;
			content_type=tok+14;
			if((p=strstr(content_type,"charset=")))
				Indexer->charset = UdmGetCharset(p + 8);
			/* Store content_type right now  */
			/* It can be modified  after     */
			/* possible external parser call */
			UdmDBEscStr(Indexer->Conf->DBType,content_type_escaped,content_type);
		}else
		if(!UDM_STRNCASECMP(tok,"Location: ")){
			location=tok+10;
		}else
		if(!UDM_STRNCASECMP(tok,"Content-Length: ")){
			size=atoi(tok+16);
		}else
		if(!UDM_STRNCASECMP(tok,"Subject: ")){
#ifdef NEWS_EXT
			char hs[UDM_MAXSUBJSIZE];
			UDM_STRNCPY(hs,tok+9);
			rfc1522_decode(HeaderSubject,hs);
#endif
			UDM_STRNCPY(title,tok+9);
			rfc1522_decode(subj,title);
			strcpy(title,subj);
		}else
		if(!UDM_STRNCASECMP(tok,"From: ")){
			char fromh[UDM_MAXKEYWORDSIZE];
#ifdef NEWS_EXT
			char hf[UDM_MAXFROMSIZE];
			UDM_STRNCPY(hf,tok+6);
			rfc1522_decode(HeaderFrom,hf);
#endif
			UDM_STRNCPY(fromh,tok+6);
			rfc1522_decode(from,fromh);
			
		}else
		if(!UDM_STRNCASECMP(tok,"Newsgroups: ")){
			UDM_STRNCPY(keywords,tok+12);
#ifdef NEWS_EXT
			strncpy(HeaderGroup,tok+12,UDM_MAXGROUPSIZE);
#endif
		}else
		if(!UDM_STRNCASECMP(tok,"Date: ")){
			last_mod_time=UdmHttpDate2Time_t(tok+6);
#ifdef NEWS_EXT
			strncpy(HeaderDate,tok+6,UDM_MAXDATESIZE);
#endif
		}else
#ifdef NEWS_EXT
		if(!UDM_STRNCASECMP(tok,"References: ")){
			strncpy(HeaderRefs,tok+12,UDM_MAXREFSIZE);
		}else
#endif
		if(!UDM_STRNCASECMP(tok,"Last-Modified: ")){
			last_mod_time=UdmHttpDate2Time_t(tok+15);
		}
		UdmLog(Indexer,UDM_LOG_DEBUG,"%s",tok);
		tok=UdmGetToken(NULL,"\r\n",&lt);
	}
	UdmLog(Indexer,UDM_LOG_EXTRA,"%s %s %d",statusline?statusline:"?",content_type?content_type:"?",size);


#ifdef NEWS_EXT
	/* do correct string termination */
	HeaderDate[UDM_MAXDATESIZE-1] = 0;
	HeaderFrom[UDM_MAXFROMSIZE-1] = 0;
	HeaderSubject[UDM_MAXSUBJSIZE-1] = 0;
	HeaderGroup[UDM_MAXGROUPSIZE-1] = 0;
	HeaderRefs[UDM_MAXREFSIZE-1] = 0;
	if((SQLDate=UdmParseDate(HeaderDate))){
		UdmDBEscStr(Indexer->Conf->DBType,HeaderDateEsc,UdmParseDate(HeaderDate));
		free(SQLDate);
	}
	/* Escape Headers */
	UdmDBEscStr(Indexer->Conf->DBType,HeaderFromEsc,HeaderFrom);
	UdmDBEscStr(Indexer->Conf->DBType,HeaderSubjectEsc,HeaderSubject);
	UdmDBEscStr(Indexer->Conf->DBType,HeaderGroupEsc,HeaderGroup);
	UdmDBEscStr(Indexer->Conf->DBType,HeaderRefsEsc,HeaderRefs);
#endif

	switch(UdmHTTPResponseType(status)){
	case 1: /* No HTTP code */
		UdmAddServerErr(Indexer->Conf,realURL.hostname);
		UdmLog(Indexer,UDM_LOG_ERROR,"No HTTP response status");
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->net_error_delay_time);
		break;

	case UDM_HTTP_STATUS_OK:
	case UDM_HTTP_STATUS_PARTIAL_OK:
		if(!content_type){
			UdmLog(Indexer,UDM_LOG_ERROR,"No Content-type in '%s'!",Doc->url);
	                UdmAddServerErr(Indexer->Conf,realURL.hostname);
			result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->net_error_delay_time);
		}
		break;

	case UDM_HTTP_STATUS_REDIRECT: /* We'll try to use Location: xxx instead */

		if((Doc->hops<CurSrv->maxhops)&&(location)){
			int newMethod;
			newMethod=UdmFilterType(Indexer->Conf,location,reason);
			if(newMethod!=UDM_DISALLOW){
				UDM_URL	newURL;
				int res;
				if(!(res=UdmParseURL(&newURL,location))){
					if(UdmFindServer(Indexer->Conf,location)){
						if((!CurSrv->delete_no_server)&&(!CurSrv->outside)){
							/* compare hostinfo in some cases */
							if(!strcmp(curURL.hostinfo,newURL.hostinfo)){
								if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
								if(UdmAddHref(Indexer->Conf,location,Doc->url_id,Doc->hops+1,0))
									Indexer->Conf->have_targets=1;
								if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
							}
						}else{
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
							if(UdmAddHref(Indexer->Conf,location,Doc->url_id,Doc->hops+1,0))
								Indexer->Conf->have_targets=1;
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
						}
					}
				}else{
					switch(res){
					case UDM_PARSEURL_LONG:
						UdmLog(Indexer,UDM_LOG_ERROR,"Redirect URL too long: '%s'",location);
						break;
					case UDM_PARSEURL_BAD:
					default:
						UdmLog(Indexer,UDM_LOG_ERROR,"Error in redirect URL: '%s'",location);
					}
				}
			}
		}
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case UDM_HTTP_STATUS_NOT_MODIFIED:  /* Not Modified, nothing to do */
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case UDM_HTTP_STATUS_DELETE:
		/* delete it if not robots.txt */
		UdmLog(Indexer,UDM_LOG_EXTRA,"Deleting URL");
		if(!strcmp(curURL.filename,"robots.txt")){
			result=UdmDeleteRobotsFromHost(Indexer,curURL.hostinfo);
			if(result==IND_OK)result=UdmLoadRobots(Indexer);
		}
		if(result!=IND_ERROR){
			if(CurSrv->deletebad){
				result=UdmDeleteUrl(Indexer,Doc->url_id);
			}else{
				result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
			}
		}
		break;

	case UDM_HTTP_STATUS_RETRY: /* We'll retry later, maybe host is down */
                UdmAddServerErr(Indexer->Conf,realURL.hostname);
		UdmLog(Indexer,UDM_LOG_EXTRA,"Could not read URL");
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->net_error_delay_time);
		break;

	default: /* Unknown status, retry later */
		UdmLog(Indexer,UDM_LOG_WARN,"HTTP %d We don't yet know how to handle it, skipped",status);
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}

	/* Check again whether "terminate" pressed */
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;

	/* Return if Content parsing is not required */
	if(result){
		FreeDoc(Doc);
		UDM_FREE(hcopy);
		return(result);
	}

	/* Now we have HTTP_OK and know Content-Type */
	/* Lets start parse the body                 */

	/* Put into mirror if required */
	if ((CurSrv->use_mirror>=0)&&(!found_in_mirror)){
		char errstr[UDMSTRSIZ]="";
		/* I tried to get it from mirror, but nothing was there  */
		/* or it was expired. Now we will store it on the mirror */
		if(UdmMirrorPUT(Indexer,CurSrv->mirror_root,CurSrv->mirror_headers,
				realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				hcopy, content, size,errstr)){
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s",errstr);
		}
	}
	UDM_FREE(hcopy);


#ifdef USE_PARSER
	/* Let's try to start external parser for this Content-Type */
	if((mime=UdmExecParser(Indexer->Conf,content_type,&mimeno,content,realsize,Indexer->Conf->max_doc_size-(content-Indexer->buf)))){
		content=mime;
		content_type=UdmParserParam(Indexer->Conf,mimeno,UDM_MIME_TO);
		if((mime=UdmParserParam(Indexer->Conf,mimeno,UDM_MIME_CHARSET)))
			Indexer->charset=UdmGetCharset(mime);
		UdmLog(Indexer,UDM_LOG_DEBUG,"Parser-Content-Type: %s",
			content_type?content_type:"unknown");
#ifdef DEBUG_PARSER
		fprintf(stderr,"content='%s'\n",content);
#endif
	};
#endif

	/* robots.txt */
	if(!UDM_STRNCASECMP(content_type,"text/plain")&&
	(!strcmp(curURL.filename,"robots.txt"))){
		result=UdmParseRobots(Indexer,content,curURL.hostinfo);
		if(result!=IND_ERROR)result=UdmLoadRobots(Indexer);
		if(result==IND_ERROR){
			FreeDoc(Doc);
			return(IND_ERROR);
		}
	}else

	/* plain TEXT or the same */
	if(!UDM_STRNCASECMP(content_type,"text/plain")||
		!UDM_STRNCASECMP(content_type,"text/tab-separated-values")||
		!UDM_STRNCASECMP(content_type,"text/css"))
	{
		if(Indexer->charset!=Indexer->Conf->local_charset){
			UdmRecode(content,Indexer->charset,Indexer->Conf->local_charset);
		} else {
			int DCindex=0;
#ifdef USE_CHARSET_GUESSER
			DCindex = UdmGuessCharset(content, UdmGetCharset(CurSrv->charset));
#else
			DCindex = UdmGetCharset(CurSrv->charset);
#endif
			UdmRecode(content,DCindex,Indexer->Conf->local_charset);
		}
		if(Method!=UDM_HEAD){
			crc32=UdmCRC32(content, realsize);
			changed=!(crc32==Doc->crc32);
			if(CurSrv->use_clones){
				origin=UdmFindOrigin(Indexer, crc32, size);
				origin=((origin==Doc->url_id)?0:origin);
			}
		}
		if(((index)&&(!origin)&&(changed))||reindex){
			strncpy(text,content,UDM_MAXTEXTSIZE-2);
			text[UDM_MAXTEXTSIZE-1]=0;
			ParseText(Indexer,CurSrv,content,CurSrv->bodyweight,1);
		}
	}else

	/* HTML text */
	if(!UDM_STRNCASECMP(content_type,"text/html")){

		if(Method!=UDM_HEAD){
			crc32=UdmCRC32(content, realsize);
			changed=!(crc32==Doc->crc32);
			if(CurSrv->use_clones){
				origin=UdmFindOrigin(Indexer, crc32, size);
				origin=((origin==Doc->url_id)?0:origin);
			}
		}

		if(((index||follow)&&(!origin)&&(changed))||reindex){
			ParseHtml(Indexer,CurSrv,content,&curURL,
				Doc,index,follow,text,keywords,descript,title);
		}
	}else{
		/* Unknown Content-Type */
		index=0; /* What this line for? */
		if(Method!=UDM_HEAD){
			/*
			MD5_DIGEST(content,realsize,digest);
			changed=strcmp(digest,Doc->crc);
			*/
			crc32=UdmCRC32(content, realsize);
			changed=!(crc32==Doc->crc32);

			if(CurSrv->use_clones){
				origin=UdmFindOrigin(Indexer, crc32, size);
				origin=((origin==Doc->url_id)?0:origin);
			}
		}
	}

	if (strcmp(curURL.filename,"robots.txt")) {
		char str[UDMSTRSIZ];
		if(CurSrv->urlweight) {
			strcpy(str,Doc->url);
			ParseText(Indexer,CurSrv,str,CurSrv->urlweight,1);
		} else {
			if(CurSrv->urlhostweight) {
				strcpy(str,curURL.hostname);
				ParseText(Indexer,CurSrv,str,CurSrv->urlhostweight,1);	
			}
			if(CurSrv->urlpathweight) {
				strcpy(str,curURL.path);
				ParseText(Indexer,CurSrv,str,CurSrv->urlpathweight,1);
			}
			if(CurSrv->urlfileweight) {
				strcpy(str,curURL.filename);
				ParseText(Indexer,CurSrv,str,CurSrv->urlfileweight,1);
			}
		}   
	}

	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Updating",Doc->url);
	if(origin){
		UdmLog(Indexer,UDM_LOG_EXTRA,"Duplicate Document with #%d",origin);
		result=UdmDeleteWordFromURL(Indexer,Doc->url_id);
		if(result==IND_OK)
			result=UdmUpdateClone(Indexer,Doc->url_id,status,CurSrv->period,
				content_type_escaped,last_mod_time,crc32);
	}else
	if((!changed)&&(!reindex)){
		result=UdmUpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}else{
		if(subj[0]){
			char str[UDMSTRSIZ];
			strcpy(str,subj);
			ParseText(Indexer,CurSrv,str,CurSrv->titleweight,1);
		}
		if((result=UdmStoreWords(Indexer,Doc->url_id,CurSrv->category?CurSrv->category:"",Doc->status))==IND_OK){
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			if(UdmStoreHrefs(Indexer))
				Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
			if(UdmDBErrorCode(Indexer->db))
				result=IND_ERROR;
			else
				result=IND_OK;
		}
		if(result!=IND_ERROR){
			int total,i;
			char str[UDMSTRSIZ]="";
			char lang[3]="";
			char *s;

			/* Prepare text,keywords,description */
			if(!strcasecmp(realURL.schema,"news"))strcpy(keywords,from);

			s=text;
#ifndef NEWS_EXT
			while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
#endif
			text[UDM_MAXTEXTSIZE-1]=0;
			UdmDBEscStr(Indexer->Conf->DBType,text_escaped,text);

			s=keywords;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			keywords[UDM_MAXKEYWORDSIZE-1]=0;
			UdmDBEscStr(Indexer->Conf->DBType,keywords_escaped,keywords);

			s=descript;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			descript[UDM_MAXDESCSIZE-1]=0;
			UdmDBEscStr(Indexer->Conf->DBType,descript_escaped,descript);

			s=title;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			title[UDM_MAXTITLESIZE-1]=0;
			UdmDBEscStr(Indexer->Conf->DBType,title_escaped,title);

			/* Guess the language */
			qsort((void*)(Indexer->lang+1),UDM_LANGPERDOC-1,sizeof(UDM_LANG),cmplang);
			total=0;
			for(i=1;i<UDM_LANGPERDOC;i++)
				total+=Indexer->lang[i].count;
			if(total){
				sprintf(str,"total: %d",total);
				for(i=1;i<UDM_LANGPERDOC;i++){
					if(Indexer->lang[i].count)
						sprintf(UDM_STREND(str)," %s:%d (%d%%)",
						Indexer->lang[i].lang,
						Indexer->lang[i].count,
						(int)(Indexer->lang[i].count*100/total));
				}
				UdmLog(Indexer,UDM_LOG_DEBUG,"language: %s",str);
				if(Indexer->lang[1].count>5){
					strncpy(lang,Indexer->lang[1].lang,2);
					lang[2]=0;
				}
			}
#ifdef NEWS_EXT
			/*
			extract message id from url
			thanks guys, the message id is in realURL.filename !!
			valid message ids have an @ character
			*/

			if(strchr(realURL.filename,'@'))
				UdmDBEscStr(Indexer->Conf->DBType,MessageIdEsc,realURL.filename);
						
			/* get rec_id from my parent out of db (if I have one...) */

			if(strlen(HeaderRefs))
			{
				/* HeaderRefs contains all message ids of my predecessors, space separated*/
				/* my direct parent is the last in the list*/
				if((parent = strrchr(HeaderRefs,' ')))
				{	
					/* parent now points to the space character */
					/* skip it */
					++parent;
				}
				else
				{
					/* there is only one entry in HeaderRefs, so this is my parent */
					parent=HeaderRefs;	
				}	
			}
			result = 0;
			/* if the parent is really a message id, */
			/* get its rec_id from database          */
			if(parent && strlen(parent) && strchr(parent,'@'))
				result = UdmFindMsgID(Indexer,parent);	
			/*
			now register me with my parent
			result is -1 if no parent was found
			*/
#ifdef HEIKODEBUG
			fprintf(stderr,"%s, me: %d, parent: %d\n",parent,Doc->url_id,result);
#endif
			if(result > 0)
				result = UdmRegisterChild(Indexer,result,Doc->url_id);

			result=UdmLongUpdateUrl(Indexer,Doc->url_id,status,
				changed,size,CurSrv->period,
				CurSrv->tag?CurSrv->tag:"",
				index,last_mod_time,
				text_escaped,title_escaped,
				content_type_escaped,keywords_escaped,
				descript_escaped,crc32,lang,
				CurSrv->category?CurSrv->category:"",
				HeaderDateEsc,
				HeaderSubjectEsc,
				HeaderFromEsc,
				HeaderGroupEsc,
				HeaderRefsEsc,
				MessageIdEsc);
#else
			result=UdmLongUpdateUrl(Indexer,Doc->url_id,status,
				changed,is_mp3?mp3_size:size,CurSrv->period,
				CurSrv->tag?CurSrv->tag:"",
				index,last_mod_time,
				text_escaped,title_escaped,
				content_type_escaped,keywords_escaped,
				descript_escaped,crc32,lang,
				CurSrv->category?CurSrv->category:"");
#endif			
		}
	}
	FreeWords(Indexer);
	FreeDoc(Doc);
	return(result);
}


/************ Misc functions *****************************/


__INDLIB__ UDM_AGENT * UdmAllocAgent(UDM_ENV * Env, int handle, int db_open_mode){
	UDM_AGENT * result;
	int i;

	result=(UDM_AGENT*)malloc(sizeof(UDM_AGENT));
	memset(result, 0, sizeof(UDM_AGENT));
	result->Conf=Env;
	result->mwords=0;
	result->nwords=0;
	result->swords=0;
	result->handle=handle; /* Handle is used in multi-threaded version */
	result->action=0;
	result->state[0]=0;
	result->Word=NULL;
	result->buf=NULL;
	result->charset=UDM_CHARSET_USASCII;
	result->db=UdmAllocDB(result, db_open_mode);
        /* Reserve connection structure */
        result->connp = (UDM_CONN*)UdmXmalloc(sizeof(UDM_CONN));
        result->connp->indexer = result;
        result->connp->connp = (UDM_CONN*)UdmXmalloc(sizeof(UDM_CONN));
        result->connp->connp->indexer = result;

	result->wordinfo[0]=0;
	result->search_mode=0;
	result->sort_order=UDM_ORD_RATE;
	result->page_number=0;
	result->page_size=0;
	result->total_found=0;
	result->group_mask=0xffffffff;
	result->track_mode=0;
	UDMMEMZERO(&result->t_Spell,sizeof(UDM_SPELL));
	result->read_timeout=300;
	result->maxsize=1024*1024;
	result->cache_mode=UDM_CACHE_DISABLED;
	result->work_time=0;
	for(i=0;i<257;i++){
		result->crcdict[i]=-1;
	}

	return(result);
}
__INDLIB__ void UdmFreeAgent(UDM_AGENT *Indexer){

	if(Indexer){
		FreeWords(Indexer);
		UDM_FREE(Indexer->Word);
		UDM_FREE(Indexer->buf);
		UDM_FREE(Indexer->connp->connp);
		UDM_FREE(Indexer->connp);
		UdmFreeDB(Indexer);
		free(Indexer);
	}
}
__INDLIB__ int UdmClearDatabase(UDM_AGENT * Indexer){
	return(UdmClearDB(Indexer));
}

__INDLIB__ char * UdmAgentErrorMsg(UDM_AGENT *Indexer){
	return(UdmDBErrorMsg(Indexer->db));
}

__INDLIB__ void UdmSetAction(UDM_AGENT *Indexer,int action){
	Indexer->action=action;
}


__INDLIB__ void UdmFreeResult(UDM_DOCUMENT * r){
	int i;
	
	if(!r)return;
	for(i=0;r[i].url_id;i++){
		UDM_FREE(r[i].url);
		UDM_FREE(r[i].content_type);
		UDM_FREE(r[i].title);
		UDM_FREE(r[i].text);
		UDM_FREE(r[i].keywords);
		UDM_FREE(r[i].description);
	}
	UDM_FREE(r);
	return;
}
