#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#if (WIN32|WINNT)
#include <winsock.h>
#include <time.h>
#include <direct.h>
#else
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#ifdef HAVE_SELECT_H
#include <select.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <regex.h>

#include "udm_common.h"
#include "udm_proto.h"
#include "udm_utils.h"
#include "udm_mutex.h"
#include "udm_socket.h"
#include "udm_http.h"
#include "udm_ftp.h"
#include "udm_xmalloc.h"
#include "udm_host.h"


#ifndef INADDR_NONE
#define INADDR_NONE (unsigned long)(-1)
#endif


/********* AddType stuff *********************/

typedef struct mime_struct {
	regex_t	reg;
	char	*mime_type;
} UDM_MIME;

int UdmAddType(UDM_ENV * Conf,char * mime_type,char *reg, char *errstr){
#define ERRSTRSIZE 100
char regerrstr[ERRSTRSIZE];
int err;
UDM_MIME * Mime;
	if(Conf->nmimes>=Conf->mmimes){
		if(Conf->mmimes){
			Conf->mmimes+=16;
			Conf->Mime=realloc(Conf->Mime,Conf->mmimes*sizeof(UDM_MIME));
		}else{
			Conf->mmimes=16;
			Conf->Mime=malloc(Conf->mmimes*sizeof(UDM_MIME));
		}
	}
	Mime=Conf->Mime;
	Mime[Conf->nmimes].mime_type=strdup(mime_type);
	err=regcomp(&(Mime[Conf->nmimes].reg),reg,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err,&(Mime[Conf->nmimes].reg),regerrstr,ERRSTRSIZE);
		sprintf(errstr,"Wrong regex in config file: %s: %s", reg,regerrstr);
		free(Mime[Conf->nmimes].mime_type);
		return(1);
	}
	Conf->nmimes++;
	return(0);
}
void UdmFreeTypes(UDM_ENV * Conf){
int i;
UDM_MIME * Mime;
	Mime=(UDM_MIME*)Conf->Mime;
	for(i=0;i<Conf->nmimes;i++){
		free(Mime[i].mime_type);
		regfree(&(Mime[i].reg));
	}
	Conf->nmimes=Conf->mmimes=0;
	UDM_FREE(Conf->Mime);
}

static char * ContentType(UDM_ENV * Conf,char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
UDM_MIME * Mime;
	Mime=Conf->Mime;
	for(i=0;i<Conf->nmimes;i++){
		err=regexec(&(Mime[i].reg),url,NS,subs,0);
		if(!err)return(Mime[i].mime_type);
	}
	return(NULL);
}

/**************************************************/

int UdmBuild(char *path, int omode) {
	struct stat sb;
	mode_t numask, oumask;
	int first, last, retval;
	char *p;

	p = path;
	oumask = 0;
	retval = 0;
	if (p[0] == UDMSLASH)		/* Skip leading slashes. */
		++p;
	for (first = 1, last = 0; !last ; ++p) {
		if (p[0] == '\0')
			last = 1;
		else if (p[0] != UDMSLASH)
			continue;
		*p = '\0';
		if (p[1] == '\0')
			last = 1;
		if (first) {
			/*
			 * POSIX 1003.2:
			 * For each dir operand that does not name an existing
			 * directory, effects equivalent to those cased by the
			 * following command shall occcur:
			 *
			 * mkdir -p -m $(umask -S),u+wx $(dirname dir) &&
			 *    mkdir [-m mode] dir
			 *
			 * We change the user's umask and then restore it,
			 * instead of doing chmod's.
			 */
			oumask = umask(0);
#if (WIN32||WINNT)
			numask = oumask;
#else
			numask = oumask & ~(S_IWUSR | S_IXUSR);
#endif
			(void)umask(numask);
			first = 0;
		}
		if (last)
			(void)umask(oumask);
		if (stat(path, &sb)) {
			if (errno != ENOENT ||
#if (WIN32||WINNT)
			    mkdir(path) < 0
#else
			    mkdir(path, last ? omode :
				  S_IRWXU | S_IRWXG | S_IRWXO) < 0
#endif
				 ){
				/* warn("%s", path); */
				retval = 1;
				break;
			}
		}
		else if ((sb.st_mode & S_IFMT) != S_IFDIR) {
			if (last)
				errno = EEXIST;
			else
				errno = ENOTDIR;
			/* warn("%s", path); */
			retval = 1;
			break;
		}
		if (!last)
			*p = UDMSLASH;
	}
	if (!first && !last)
		(void)umask(oumask);
	return (retval);
}


/*************** Mirroring stuff ************************/

int   UdmMirrorGET(UDM_AGENT * Indexer,
		char * MirrorRoot,
		char * MirrorHeadersRoot,
		char *schema, char *hostname, char *path,
		char *filename, char *buf, int maxsize,
		int days, char *errstr) {

	int size = 0, l;
	int fbody, fheader;
	char str[UDMSTRSIZ]="";
	char *content_type = NULL;
	struct stat sb;
	time_t nowtime;
	int have_headers=0;

	nowtime = time(NULL);

	if (days <= 0)return(0);
	
	/* MirrorRoot is not specified, nothing to do */
	if(!MirrorRoot)return(0);

	strcpy(str, MirrorRoot);strcat(str,"/");
	strcat(str, schema);strcat(str,"/");
	strcat(str, hostname);
	strcat(str, path);
	strcat(str, filename && strlen(filename) ? filename : "index.html");

	if ((fbody = open(str, O_RDONLY)) == -1){
		sprintf(errstr,"Mirror file %s not found", str);
		return UDM_MIRROR_NOT_FOUND;
	}

	/* Check on file mtime > days ? return */
	if (fstat(fbody, &sb)) {
		/* error handler */
	}
	if (nowtime > sb.st_mtime + days ) {
		close(fbody);
		sprintf(errstr,"%s is older then %d secs, retrieving", str, days);
		return UDM_MIRROR_EXPIRED;
	}

	if(MirrorHeadersRoot){
		strcpy(str, MirrorHeadersRoot);strcat(str,"/");
		strcat(str, schema);strcat(str,"/");
		strcat(str, hostname);
		strcat(str, path);
		strcat(str, filename && strlen(filename) ? filename : "index.html");
		strcat(str,".header");

		if ((fheader = open(str, O_RDONLY))>=0) {
			size = read(fheader, buf, maxsize);
			close(fheader);
			strcpy(buf + size, "\r\n\r\n");
			have_headers=1;
		}
	}
	if(!have_headers){
		/* header file not found   */
		/* get body as file method */
		sprintf(buf,"HTTP/1.0 200 OK\r\n");
		if((content_type=ContentType(Indexer->Conf,filename && strlen(filename) ? filename : "index.html")))
			sprintf(UDM_STREND(buf),"Content-Type: %s\r\n",content_type);
		sprintf(UDM_STREND(buf),"\r\n");
	}

	l = strlen(buf);
	size = read(fbody, buf + l, maxsize - l) + l;
	close(fbody);
	return size;
}

int   UdmMirrorPUT(UDM_AGENT * Indexer,
		char * MirrorRoot, char * MirrorHeadersRoot,
		char *schema, char *hostname, char *path, char *filename, 
		char *header, char *content, int size, char *errstr) {

	int fd;
	char str[UDMSTRSIZ]="";

	/* Put Content if MirrorRoot is specified */
	if(MirrorRoot){
		strcpy(str, MirrorRoot);strcat(str,"/");
		strcat(str, schema);strcat(str,"/");
		strcat(str, hostname);
		strcat(str, path);

		if(UdmBuild(str, 0755) != 0){
			sprintf(errstr, "Can't create dir %s", str);
			return UDM_MIRROR_CANT_BUILD;
		}

		strcat(str,"/");
		strcat(str, filename && strlen(filename) ? filename : "index.html");
		if ((fd = open(str, O_CREAT|O_WRONLY,0644)) == -1){
			sprintf(errstr,"Can't open mirror file %s\n", str);
			return UDM_MIRROR_CANT_OPEN;
		}
		size = write(fd, content, size);
		close(fd);
	}

	/* Put Headers if MirrorHeadersRoot is specified */
	if(MirrorHeadersRoot){
		strcpy(str, MirrorHeadersRoot);strcat(str,"/");
		strcat(str, schema);strcat(str,"/");
		strcat(str, hostname);
		strcat(str, path);

		if(UdmBuild(str, 0755) != 0){
			sprintf(errstr, "Can't create mirror dir %s", str);
			return UDM_MIRROR_CANT_BUILD;
		}

		strcat(str,"/");
		strcat(str, filename && strlen(filename) ? filename : "index.html");
		strcat(str,".header");

		if ((fd = open(str, O_CREAT|O_WRONLY,0644)) == -1){
			sprintf(errstr,"Can't open mirror file %s\n", str);
			return UDM_MIRROR_CANT_OPEN;
		}
		size = write(fd, header, strlen(header));
		close(fd);
	}
	return(0);
}


/************** Open TCP Connect **************/

static int open_host(UDM_AGENT *Indexer, char *hostname,int port, int timeout)
{
int net, len;

	UDMMEMZERO((char*)&Indexer->connp->sin, sizeof(struct sockaddr_in));
	if (!port)
		return(UDM_NET_ERROR);
	Indexer->connp->port = port;
	Indexer->connp->sin.sin_family=AF_INET;

        len = strlen(hostname);
        Indexer->connp->hostname = UdmXrealloc(Indexer->connp->hostname, len+1);
        udm_snprintf(Indexer->connp->hostname, len+1, "%s", hostname);

	if (UdmHostLookup(Indexer->Conf,Indexer->connp)==-1)
		return Indexer->connp->err;

	net=socket(AF_INET, SOCK_STREAM, 0);

	if(connect(net, (struct sockaddr *)&Indexer->connp->sin, sizeof (struct sockaddr_in)))
		return(UDM_NET_CANT_CONNECT);


	return(net);
}


/**************** HTTP codes and messages ************/

char *UdmHTTPErrMsg(int code){
	switch(code){
	case 0:   return("Not indexed yet");
	case 200: return("OK");
	case 206: return("Partial OK");
	case 301: return("Moved Permanently");
	case 302: return("Moved Temporarily");
	case 303: return("See Other");
	case 304: return("Not Modified");
	case 300: return("Multiple Choices");
	case 305: return("Use Proxy (proxy redirect)");
	case 400: return("Bad Request");
	case 401: return("Unauthorized");
	case 402: return("Payment Required");
	case 403: return("Forbidden");
	case 404: return("Not found");
	case 405: return("Method Not Allowed");
	case 406: return("Not Acceptable");
	case 407: return("Proxy Authentication Required");
	case 408: return("Request Timeout");
	case 409: return("Conflict");
	case 410: return("Gone");
	case 411: return("Length Required");
	case 412: return("Precondition Failed");
	case 413: return("Request Entity Too Large");
	case 414: return("Request-URI Too Long");
	case 415: return("Unsupported Media Type");
	case 500: return("Internal Server Error");
	case 501: return("Not Implemented");
	case 502: return("Bad Gateway");
	case 505: return("Protocol Version Not Supported");
	case 503: return("Service Unavailable");
	case 504: return("Gateway Timeout");
	default:  return("Unknown status");
	}
}

int UdmHTTPResponseType(int status){

	switch(status){
	case 1:   return(1);
	case 200: return(UDM_HTTP_STATUS_OK);
	case 206: return(UDM_HTTP_STATUS_PARTIAL_OK);

	case 301:  /* Moved Permanently */
	case 302:  /* Moved Temporarily */
	case 303:  /* See Other */ return(UDM_HTTP_STATUS_REDIRECT);

	case 304:  /* Not Modified*/ return(UDM_HTTP_STATUS_NOT_MODIFIED);

	case 300:  /* Multiple Choices              */
	case 305:  /* Use Proxy (proxy redirect)    */
	case 400:  /* Bad Request ??? We tried ...  */
	case 401:  /* Unauthorized                   */
	case 402:  /* Payment Required              */
	case 403:  /* Forbidden                     */
	case 404:  /* Not found                     */
	case 405:  /* Method Not Allowed            */
	case 406:  /* Not Acceptable                */
	case 407:  /* Proxy Authentication Required */
	case 408:  /* Request Timeout               */
	case 409:  /* Conflict                      */
	case 410:  /* Gone                          */
	case 411:  /* Length Required               */
	case 412:  /* Precondition Failed           */
	case 413:  /* Request Entity Too Large      */
	case 414:  /* Request-URI Too Long          */
	case 415:  /* Unsupported Media Type        */
	case 500:  /* Internal Server Error         */
	case 501:  /* Not Implemented               */
	case 502:  /* Bad Gateway                   */
	case 505:  /* Protocol Version Not Supported */
		return(UDM_HTTP_STATUS_DELETE);

	case 503: /* Service Unavailable */
	case 504: /* Gateway Timeout */
		return(UDM_HTTP_STATUS_RETRY);

	default: return(UDM_HTTP_STATUS_UNKNOWN);
	}
}


/*******************************************************/
#ifdef USE_HTTP

int UdmHTTPGetRange(UDM_AGENT *Indexer,char *header,char *host,int port,char *range)
{
    char *range_header;
    int size;

	    if (http_connect(Indexer->Conf,Indexer->connp, host, port, Indexer->read_timeout)==-1)
		    return -1;
	    /* make Range header */
	    range_header=UdmXmalloc(strlen(header)+ 17 +strlen(range));
	    udm_snprintf(range_header, strlen(header)-1, "%s", header);
	    sprintf(UDM_STREND(range_header), "Range: bytes=%s\r\n\r\n", range);
	    if (socket_write(Indexer->connp, range_header)==-1){
		    UDM_FREE(range_header);
		    return -1;
	    }
	    UDM_FREE(range_header);
	    if ((size =socket_read(Indexer->connp, Indexer->maxsize))==-1)
		    return -1;

	    if (size > Indexer->maxsize)
		    size = Indexer->maxsize;

	    memcpy(Indexer->buf, Indexer->connp->buf, size);
	    UDM_FREE(Indexer->connp->buf);
	    socket_close(Indexer->connp);
	    return size;
}

int UdmHTTPGet(UDM_AGENT *Indexer,char * header,char *host,int port)
{
    int fd;
    fd_set sfds;
    int status, nread=0;

    struct timeval tv;

    /* Connect to HTTP or PROXY server */
    if( (fd=open_host(Indexer,host,port,Indexer->read_timeout)) < 0 )
       return(fd);

    /* Send HTTP request */
    if( send(fd,header,strlen(header),0) < 0 )
       return -1;

    /* Retrieve response */
    tv.tv_sec = (long) Indexer->read_timeout;
    tv.tv_usec = 0;

    while( 1 ){
       FD_ZERO( &sfds );
       FD_SET( fd, &sfds );

       status=select(FD_SETSIZE, &sfds, NULL, NULL, &tv);
       if( status == -1 ){
               nread=UDM_NET_ERROR;
               break;
       }else if( status == 0 ){
               nread=UDM_NET_TIMEOUT;
               break;
       }else{
           if(FD_ISSET(fd,&sfds))
           {
               status = recv(fd, Indexer->buf+nread, NET_BUF_SIZE, 0);
               if( status < 0 )
               {
                   nread = status;
                   break;
               }else if( status == 0 ){
                   nread += status;
                   break;
               }else{
                   nread += status;
                   if( (nread+NET_BUF_SIZE) >= Indexer->maxsize )
                       break;
               }
           }else
               break;
       }
    }
    close(fd);
    return(nread);
}
#endif

#ifdef USE_FTP
int UdmFTPGet(UDM_AGENT * Indexer,char *host,int port, 
	char *path, char *filename,
	char *user, char *passwd,
	int last_mod_time, int checkonly)
{
	int size=0, last_mod_tm, buf_size, code, len;
	char *con_type, buf[256], *full_path=NULL;
	
	/* Check for changing host */
	if (!Indexer->connp->hostname || 
	    strcmp(Indexer->connp->hostname, host) || 
	    (Indexer->connp->connected == NET_NOTCONNECTED)){
		code=ftp_connect(Indexer->Conf,Indexer->connp, host, port, user, passwd, Indexer->read_timeout);
		if (code == -1){
			if (Indexer->connp->err >0){
				sprintf(Indexer->buf, "HTTP/1.1 401 OK\r\n\r\n  ");
				size = strlen(Indexer->buf);
			}else{
				size = Indexer->connp->err;
			}
		}
	}
	
	if (Indexer->connp->connected == NET_CONNECTED){
		/* Make listing */
		if (!filename){
			code = ftp_cwd(Indexer->connp, path);
			if (code != -1){
				code = ftp_list(Indexer->connp, Indexer->connp->connp, path, NULL, Indexer->maxsize);
				if (code == -1){
					if (Indexer->connp->err >0){
						sprintf(Indexer->buf, "HTTP/1.1 403 OK\r\n\r\n");
						size = strlen(Indexer->buf);
					}else{
						size = Indexer->connp->err;
					}
				}else{
					udm_snprintf(Indexer->buf, Indexer->maxsize, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n<html><body>%s</body></html>", 
						Indexer->connp->connp->buf);
					size = strlen(Indexer->buf);
				}
			}else{
				if (Indexer->connp->err >0){
					sprintf(Indexer->buf, "HTTP/1.1 403 OK\r\n\r\n");
					size = strlen(Indexer->buf);
				}else{
					size = Indexer->connp->err;
				}
			}
		}else{
			con_type = ContentType(Indexer->Conf,filename);
			len = strlen(path)+strlen(filename);
			full_path = UdmXmalloc(len+1);
			udm_snprintf(full_path, len+1, "%s%s", path, filename);
			last_mod_tm = ftp_mdtm(Indexer->connp,full_path);
			if(last_mod_tm == -1 && Indexer->connp->err){
				if (Indexer->connp->err >0){
					sprintf(Indexer->buf, "HTTP/1.1 404 OK\r\n\r\n");
					size = strlen(Indexer->buf);
				}else
					size = Indexer->connp->err;

			}else if (last_mod_tm == last_mod_time){
				sprintf(Indexer->buf, "HTTP/1.1 304 OK\r\n\r\n");
				size = strlen(Indexer->buf);
			}else{
				UdmTime_t2HttpStr(last_mod_tm, buf);
				if (!checkonly){
					if (!ftp_get(Indexer->connp, Indexer->connp->connp, full_path,Indexer->maxsize)){
						udm_snprintf(Indexer->buf, Indexer->maxsize, "HTTP/1.1 200 OK\r\nContent-Type: %s\n\rLast-Modified: %s\r\n\r\n", 
							con_type,buf);
						size = strlen(Indexer->buf);
						if (Indexer->connp->connp->buf_len+size >= Indexer->maxsize)
							buf_size = Indexer->maxsize-size;
						else
							buf_size = Indexer->connp->connp->buf_len;
						
						memcpy(Indexer->buf+size, Indexer->connp->connp->buf, buf_size);
						size +=buf_size; 
					}else{
						if (Indexer->connp->err >0){
							sprintf(Indexer->buf, "HTTP/1.1 403 OK\r\n\r\n");
							size = strlen(Indexer->buf);
						}else{
							size = Indexer->connp->err;
			
						}
					}
				}else{
				    	sprintf(Indexer->buf, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nLast-Modified: %s\r\n\r\n", buf);
					size = strlen(Indexer->buf);
				}
			}
		}
	}

	UDM_FREE(full_path);
	UDM_FREE(Indexer->connp->buf);
	UDM_FREE(Indexer->connp->connp->buf);
	return size;
}
#endif  

#ifdef USE_NEWS
int UdmNEWSGet(UDM_AGENT * Indexer,char *header,char *host,int port){
int fd,size,status;
char *s,*c,*e;
char str[UDMSTRSIZ];
FILE *f,*f1;
char filename[UDMSTRSIZ];
char command[32]="";
char proto[32]="";
int a=0,b=0,from=0,to=0,headers=0;
char *email=NULL;
char *msg_id=NULL;
char *subj=NULL;
int has_content=0;
int has_if_modified=0;
char *lt;

	sscanf(header,"%s%s%s",command,str,proto);
	if(str[0]=='/')strcpy(filename,str+1);
	else	strcpy(filename,str);
	s=UdmGetToken(header,"\r\n",&lt);
	while(s){
		if(!UDM_STRNCASECMP(s,"If-Modified-Since: "))
			has_if_modified=1;
		s=UdmGetToken(NULL,"\r\n",&lt);
	}

	/* Connect to NEWS server */
	fd=open_host(Indexer,host,port,Indexer->read_timeout);	/* Connect */
	if(fd<0)return(fd);			/* Could not connect */

	size=0;s=Indexer->buf;
	f=fdopen(fd,"r");f1=fdopen(fd,"w");
	fgets(str,sizeof(str),f); /* Read hello string */

	/* All news groups on the server */
	if((!strcmp(filename,"/"))||(!strcmp(filename,""))){
		fprintf(f1,"list\r\n");fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if((c=strchr(str,' ')))*c=0;
			sprintf(UDM_STREND(Indexer->buf),"<A HREF=\"news://%s/%s\"></A>\n",host,str);
		}
		sprintf(UDM_STREND(Indexer->buf),"</BODY></HTML>\n");
	}else /* List of the articles in one news group */
	if(!strchr(filename,'@')){
		if(filename[strlen(filename)-1]=='/')
			filename[strlen(filename)-1]='\0';
		fprintf(f1,"group %s\r\n",filename);fflush(f1);
		fgets(str,sizeof(str),f);
		sscanf(str,"%d%d%d%d",&a,&b,&from,&to);
		if(a != NNTP_GROUP_OK){
			/* Newsgroup does not exist */
			status=404;
			sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
		}else{
			fprintf(f1,"xover %d-%d\r\n",from,to);fflush(f1);
			fgets(str,sizeof(str),f);
			sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
			while(fgets(str,sizeof(str),f)){
				if(str[0]=='.')break;
				if((e=UdmGetToken(str,"\t\r\n",&lt)))
				if((subj=UdmGetToken(NULL,"\t\r\n",&lt)))
				if((email=UdmGetToken(NULL,"\t\r\n",&lt)))
				if((e=UdmGetToken(NULL,"\t\r\n",&lt)))
				if((msg_id=UdmGetToken(NULL,"\t\r\n",&lt)));
				str[0]=0;
				if(msg_id)sscanf(msg_id,"<%[^>]s>",str);
				if(str[0] && ((strlen(Indexer->buf)+strlen(host)+strlen(str)) < Indexer->maxsize - 20))
					sprintf(UDM_STREND(Indexer->buf),"<A HREF=\"news://%s/%s\"></A>\n",host,str);
				/* in the line above, I left 20 bytes of free */
				/* space at the end of the string for the     */
				/* closing string: */
			}
		}
		sprintf(UDM_STREND(Indexer->buf),"</BODY></HTML>\n");

	}else{ /* One article */
		if(has_if_modified){
			fprintf(f1,"stat <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==223){
				status=304; /* Not modified */
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}else{
				status=404; /* Not found */
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}
		}else{
			fprintf(f1,"article <%s>\r\n",filename);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==220){
				headers=1;has_content=0;
				sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\n");
				while(fgets(str,sizeof(str),f)){
					if(str[0]=='.')break;
					if(headers){
						if(!UDM_STRNCASECMP(str,"Content-Type:"))has_content=1;
						if((!strcmp(str,"\r\n"))||(!strcmp(str,"\n"))){
							headers=0;
							if(!has_content)strcat(Indexer->buf,"Content-Type: text/plain\r\n");
						}
					}
					if((int)(strlen(Indexer->buf)+strlen(str))<Indexer->maxsize)
						strcat(Indexer->buf,str);
				}
			}else{
				status=404;
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}
		}
	}
	fprintf(f1,"quit\r\n");fflush(f1);
	fclose(f);/* Close NEWS connection */
	fclose(f1);
	return(strlen(Indexer->buf));
}

/* FIXME !!! */
/* nntp:// schema does not work properly yet */

int UdmNNTPGet(UDM_AGENT * Indexer,char *header,char *host,int port){
int fd,size,status;
char *s,*c,*e;
char str[UDMSTRSIZ];
FILE *f,*f1;
char filename[UDMSTRSIZ];
char command[32]="";
char proto[32]="";
int a=0,b=0,from=0,to=0,headers=0;
int has_content=0;
int has_if_modified=0;
char *lt;

	sscanf(header,"%s%s%s",command,str,proto);
	if(str[0]=='/')strcpy(filename,str+1);
	else	strcpy(filename,str);

	if(filename[0])
		if(filename[strlen(filename)-1]=='/')
			filename[strlen(filename)-1]='\0';

	s=UdmGetToken(header,"\r\n",&lt);
	while(s){
		if(!UDM_STRNCASECMP(s,"If-Modified-Since: "))has_if_modified=1;
		s=UdmGetToken(NULL,"\r\n",&lt);
	}

	/* Connect to NNTP server */
	fd=open_host(Indexer,host,port,Indexer->read_timeout);	/* Connect */
	if(fd<0)return(fd);			/* Could not connect */

	size=0;s=Indexer->buf;
	f=fdopen(fd,"r");f1=fdopen(fd,"w");
	fgets(str,sizeof(str),f); /* Read hello string */

	/* All news groups on the server */
	if((!strcmp(filename,"/"))||(!strcmp(filename,""))){
		fprintf(f1,"list\r\n");fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if((c=strchr(str,' ')))*c=0;
			sprintf(UDM_STREND(Indexer->buf),"<A HREF=\"nntp://%s/%s\"></A>\n",host,str);
		}
		sprintf(UDM_STREND(Indexer->buf),"</BODY></HTML>\n");
	}else /* List of the articles in one news group */
	if(!strchr(filename,'/')){
		fprintf(f1,"group %s\r\n",filename);fflush(f1);
		fgets(str,sizeof(str),f);
		sscanf(str,"%d%d%d%d",&a,&b,&from,&to);
		fprintf(f1,"xover %d-%d\r\n",from,to);fflush(f1);
		fgets(str,sizeof(str),f);
		sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
		while(fgets(str,sizeof(str),f)){
			if(str[0]=='.')break;
			if((e=UdmGetToken(str,"\t\r\n",&lt)))
				sprintf(UDM_STREND(Indexer->buf),
				"<A HREF=\"nntp://%s/%s/%s\"></A>\n",
				host,filename,e);
		}
		sprintf(UDM_STREND(Indexer->buf),"</BODY></HTML>\n");
	}else{ /* One article */
		c = strchr(filename,'/');
		*c = '\0'; c++;
		fprintf(f1, "group %s\r\n", filename); fflush(f1);
		fgets(str, sizeof(str),f);
		if(has_if_modified){
			fprintf(f1,"stat %s\r\n",c);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==223){
				status=304; /* Not modified */
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}else{
				status=404; /* Not found */
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}
		}else{
			fprintf(f1,"article %s\r\n",c);fflush(f1);
			fgets(str,sizeof(str),f);sscanf(str,"%d",&a);
			if(a==220){
				headers=1;has_content=0;
				sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\n");
				while(fgets(str,sizeof(str),f)){
					if(str[0]=='.')break;
					if(headers){
						if(!UDM_STRNCASECMP(str,"Content-Type:"))has_content=1;
						if((!strcmp(str,"\r\n"))||(!strcmp(str,"\n"))){
							headers=0;
							if(!has_content)strcat(Indexer->buf,"Content-Type: text/plain\r\n");
						}
					}
					if((int)(strlen(Indexer->buf)+strlen(str))<Indexer->maxsize)
						strcat(Indexer->buf,str);
				}
			}else{
				status=404;
				sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			}
		}
	}
	fprintf(f1,"quit\r\n");fflush(f1);
	fclose(f);/* Close NNTP connection */
	fclose(f1);
	return(strlen(Indexer->buf));
}
#endif


#ifdef USE_FILE
int UdmFILEGet(UDM_AGENT * Indexer,char *header){
int fd,size,l,status=0,is_dir;
struct stat sb,sb1;
DIR * dir;
struct dirent *rec;
char filename[UDMSTRSIZ];
char newfilename[UDMSTRSIZ];
char openname[UDMSTRSIZ];
char command[32]="";
char proto[32]="";
char *content_type=NULL,*s,*lt;
time_t ims=0;

	sscanf(header,"%s%s%s",command,filename,proto);
	strcpy(newfilename,filename);
	UdmUnescapeCGIQuery(openname,newfilename);

	/* Remember If-Modified-Since timestamp */
	s=UdmGetToken(header,"\r\n",&lt);
	while(s){
		if(!UDM_STRNCASECMP(s,"If-Modified-Since: ")){
			ims=UdmHttpDate2Time_t(s+19);
		}
		s=UdmGetToken(NULL,"\r\n",&lt);
	}

	s=openname;
#if (WIN32|WINNT)
	while(*s){
		if(*s=='/')*s='\\';
		s++;
	}
	if ((strlen(openname)>3)&&(openname[strlen(openname)-1]=='\\'))
		openname[strlen(openname)-1]='\0';
#endif
	if(stat(openname,&sb)){
		switch(errno){
			case ENOENT: status=404;break; /* Not found */
			case EACCES: status=403;break; /* Forbidden*/
			default: status=500;
		}
		sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
		return(strlen(Indexer->buf));
	}

	/* If directory is given without ending "/"   */
	/* we must redirect to the same URL with "/"  */
	if((sb.st_mode&S_IFDIR)&&(filename[strlen(filename)-1]!='/')){
		status=301;
		sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\nLocation: file:%s/\r\n\r\n",status,UdmHTTPErrMsg(status),filename);
		return(strlen(Indexer->buf));
	}
	
	if(sb.st_mode&S_IFDIR){
		if((dir=opendir(openname))){
			strcpy(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n<HTML><BODY>\n");
			while((rec = readdir (dir))){
				char escaped_name[UDMSTRSIZ]="";
				char *s,*e;
				/*
				This does not work on Solaris
				is_dir=(rec->d_type==DT_DIR);
				*/
				sprintf(newfilename,"%s%s",filename,rec->d_name);
				if(stat(newfilename,&sb1)){
					/* FIXME!!! Handler should be here */
				}
				is_dir=((sb1.st_mode&S_IFDIR)>0);
				
				e=escaped_name;
				for(s=rec->d_name;*s;s++){
					if(strchr(" %&<>+[](){}/?#'\"\\;,",*s)){
						sprintf(e,"%%%X",(int)*s);
						e+=3;
					}else{
						*e=*s;
						e++;
					}
				}
				*e=0;
				sprintf(UDM_STREND(Indexer->buf),"<A HREF=\"%s%s\">%s%s</A>\n",
					escaped_name,is_dir?"/":"",
					escaped_name,is_dir?"/":"");
			}
			closedir(dir);
			strcpy(UDM_STREND(Indexer->buf),"</BODY><HTML>\n");
			return(strlen(Indexer->buf));
		}else{
			switch(errno){
				case ENOENT: status=404;break; /* Not found */
				case EACCES: status=403;break; /* Forbidden*/
				default: status=500;
			}
			sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			return(strlen(Indexer->buf));
		}
	}else{
		/* Lets compare last modification date */
		if(ims>=sb.st_mtime){
			/* Document seems to be unchanged */
			status=304;
			sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			size=strlen(Indexer->buf);
			return(size);
		}

		if((fd=open(openname,O_RDONLY))<0){
			switch(errno){
				case ENOENT: status=404;break;
				case EACCES: status=403;break;
				default: status=1;
			}
			sprintf(Indexer->buf,"HTTP/1.0 %d %s\r\n\r\n",status,UdmHTTPErrMsg(status));
			return(strlen(Indexer->buf));
		}
		sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\n");
		if((content_type=ContentType(Indexer->Conf,filename)))
			sprintf(UDM_STREND(Indexer->buf),"Content-Type: %s\r\n",content_type);

		strcpy(UDM_STREND(Indexer->buf),"Last-Modified: ");
		UdmTime_t2HttpStr(sb.st_mtime, UDM_STREND(Indexer->buf));
		strcpy(UDM_STREND(Indexer->buf),"\r\n");

		sprintf(UDM_STREND(Indexer->buf),"\r\n");
		l=strlen(Indexer->buf);
		size=read(fd,Indexer->buf+l,Indexer->maxsize-l)+l;
		close(fd);
		return(size);
	}
}
#endif
