#include "udm_common.h"
#include <stdlib.h>
#include <fcntl.h>
#include "udm_utils.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"

static const char search_cache_path[UDMSTRSIZ]=UDM_VAR_DIR "/cache/";
static const int search_cache_size=1000;

static int nbits[256] = {
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8,
};

static int bit_count(int v)
{
  return (nbits[v&0xFF]+nbits[(v>>8)&0xFF]+nbits[(v>>16)&0xFF]+nbits[(v>>24)&0xFF]);
}


static int cmpword(const void *s1,const void *s2){
int res;
	if(!(res=(((UDM_SEARCHWORD*)s2)->count)-(((UDM_SEARCHWORD*)s1)->count)))
		if(!(res=(((UDM_SEARCHWORD*)s2)->weight-(((UDM_SEARCHWORD*)s1)->weight))))
			if(!(res=(((UDM_SEARCHWORD*)s1)->url_id-(((UDM_SEARCHWORD*)s2)->url_id))));
	return(res);
}

/* Find topcount best results */
void UdmWrdTopSort(UDM_SEARCHWORD * wrd, int nwrd,int topcount){
	register int j;
	UDM_SEARCHWORD w;

#ifdef DEBUG_TOP_SORT
	fprintf(stderr,"top+1=%d nwrd=%d\n",topcount+1,nwrd);
#endif

	qsort((void*)wrd,topcount+1,sizeof(UDM_SEARCHWORD),cmpword);
	for(j=topcount;j<nwrd;j++){
		register int res;
		if(!(res=(wrd[j].count-wrd[topcount].count)))
		if(!(res=(wrd[j].weight-wrd[topcount].weight)))
		if(!(res=(wrd[topcount].url_id-wrd[j].url_id)));
#ifdef DEBUG_TOP_SORT
fprintf(stderr,"(%d,%d,%d) %d (%d,%d,%d) %d\n",
	wrd[topcount].count,wrd[topcount].weight,wrd[topcount].url_id,topcount,
	wrd[j].count,wrd[j].weight,wrd[j].url_id,j);
#endif
		if(res>0){
			/*
			memcpy(&w,&wrd[topcount],sizeof(UDM_SEARCHWORD));
			memcpy(&wrd[topcount],&wrd[j],sizeof(UDM_SEARCHWORD));
			memcpy(&wrd[j],&w,sizeof(UDM_SEARCHWORD));
			qsort((void*)wrd,topcount+1,sizeof(UDM_SEARCHWORD),cmpword);
			*/

			int l,c,r;
			
			l=0;r=topcount;
			while(l<r){
				c=(l+r)/2;
				if(!(res=(wrd[c].count-wrd[j].count)))
				if(!(res=(wrd[c].weight-wrd[j].weight)))
				if(!(res=(wrd[j].url_id-wrd[c].url_id)));
				if(res>0){
					l=c+1;
				}else{
					r=c;
				}
			}
			//fprintf(stderr,"l=%d r=%d\n",l,r);
			//for(l=0;l<=topcount;l++)
			//fprintf(stderr,"(%d,%d,%d) %d\n",
			//	wrd[l].count,wrd[l].weight,wrd[l].url_id,l);
			//fprintf(stderr,"(%d,%d,%d) %d\n",
			//	wrd[j].count,wrd[j].weight,wrd[j].url_id,j);
			//fprintf(stderr,"\n");

			w=wrd[topcount];
			memmove(&wrd[r+1],&wrd[r],(topcount-r)*sizeof(UDM_SEARCHWORD));
			wrd[r]=wrd[j];
			wrd[j]=w;
			
			/*for(l=0;l<=topcount;l++)
			fprintf(stderr,"(%d,%d,%d) %d\n",
				wrd[l].count,wrd[l].weight,wrd[l].url_id,l);
			fprintf(stderr,"(%d,%d,%d) %d\n",
				wrd[j].count,wrd[j].weight,wrd[j].url_id,j);
			fprintf(stderr,"END\n");
			*/
		}
	}
}

/***************** Results cache functions **************/

int UdmStoreToCache(UDM_AGENT * query,char * fname,UDM_SEARCHWORD * wrd, int nwrd){
	int fd;
	char info[1024]="";

	if(query->page_number<FAST_PRESORT_PAGES)qsort((void*)wrd,nwrd,sizeof(UDM_SEARCHWORD),cmpword);
#ifdef DEBUG_CACHE
		fprintf(stderr,"write to %s\n",fname);
#endif
		if((fd=open(fname,O_WRONLY|O_CREAT|O_TRUNC,0644))>=0){
#ifdef DEBUG_CACHE
			fprintf(stderr,"found:%d\n",nwrd);	
#endif
			write(fd,&nwrd,4);
			strncpy(info,query->wordinfo,sizeof(info));
			write(fd,info,sizeof(info));
			if((search_cache_size==0)||(nwrd<search_cache_size)){
				write(fd,wrd,nwrd*sizeof(UDM_SEARCHWORD));
			}else{
				write(fd,wrd,search_cache_size*sizeof(UDM_SEARCHWORD));
			}
			close(fd);
		}else{
#ifdef DEBUG_CACHE
			fprintf(stderr,"%s\n",strerror(errno));
#endif
		}
	UDM_FREE(wrd);
	return(0);
}

/************************************************************/

UDM_SEARCHWORD * UdmFindInCache(UDM_AGENT * Agent,char * query_id,int * nwrd){
char fname[UDMSTRSIZ];
char info[1024]="HZ";
int fd;
UDM_SEARCHWORD * wrd=NULL;
int bytes;

	*nwrd=0;
	if((search_cache_size!=0)&&((Agent->page_number+1)*Agent->page_size>search_cache_size))return(NULL);

	sprintf(fname,"%s",query_id);
	if((fd=open(fname,O_RDONLY))<0)return(NULL);
	if(-1==read(fd,&Agent->total_found,4)){
		close(fd);
		return(NULL);
	}
	if(-1==read(fd,info,sizeof(info))){
		close(fd);
		return(NULL);
	}
	strcpy(Agent->wordinfo,info);
#ifdef DEBUG_CACHE
	fprintf(stderr,"found: %d wordinfo: %s\n",*found,info);
#endif
	wrd=(UDM_SEARCHWORD*)malloc(Agent->page_size*sizeof(UDM_SEARCHWORD));
	if(-1==lseek(fd,Agent->page_number*Agent->page_size*sizeof(UDM_SEARCHWORD),SEEK_CUR)){
		close(fd);
		return(NULL);
	}
	if(-1==(bytes=read(fd,wrd,Agent->page_size*sizeof(UDM_SEARCHWORD)))){
		close(fd);
		return(NULL);
	}
	close(fd);
	*nwrd=bytes/sizeof(UDM_SEARCHWORD);

	return(wrd);
}

/****************************************************/

UDM_SEARCHWORD * UdmGroupByURL(UDM_AGENT * query,UDM_SEARCHWORD * wrd, int words_in_query, UDM_STACK_ITEM * items, int nitems){
	register int f=0;
	register int i,j;
	register int w_id=0;
	register int item;
	UDM_STACK_ITEM temp_items[UDM_MAXSTACK];

	memcpy(temp_items,items,nitems*sizeof(UDM_STACK_ITEM));

#ifdef USE_GROUP
	/* find first word in corresponding group*/
	for(f=0; f<query->total_found; f++){
	    	if((w_id = wrd[f].weight&query->group_mask)){
			unsigned int weight=0,group_id;
    			for(group_id=0; group_id<sizeof(query->group_mask)*8; group_id++){
				weight += (w_id >>group_id)&1;
			}
			wrd[0].weight=weight;
			wrd[0].url_id=wrd[f].url_id;
			wrd[0].count=wrd[f].count;
			break;
		}
	}
#endif
	/* Group by url_id */
	j=0;
	for(i=f+1;i<query->total_found;i++){
#ifdef USE_GROUP
		if((w_id = wrd[i].weight&query->group_mask)){
			unsigned int weight=0,group_id;
	    		for(group_id=0; group_id<sizeof(query->group_mask)*8; group_id++){
				weight += (w_id >>group_id)&1;
			}
			wrd[i].weight=weight;
		}
#else
	        w_id=1;
#endif
		if(wrd[j].url_id==wrd[i].url_id){
			wrd[j].weight+=wrd[i].weight;
			if(query->search_mode==UDM_MODE_BOOL){
				wrd[j].count|=wrd[i].count;
			}else{
				wrd[j].count+=wrd[i].count;
			}
		}else{
			if((query->search_mode==UDM_MODE_ALL)&&(wrd[j].count<words_in_query)){
				/* Skip this result */
			}else
			if(query->search_mode==UDM_MODE_BOOL){
				
				/* Create a copy of lexem array       */
				/* and change word masks by 1 or 0    */
				/* depending on whether word presents */
				/* in the search query being executed */
				for(item=0;item<nitems;item++){
					if(temp_items[item].cmd==UDM_STACK_WORD){
						/* Change word mask by 1 or 0 */
						temp_items[item].arg=((items[item].arg)&(wrd[j].count))?1:0;
					}
				}
				if(UdmCalcBoolItems(temp_items,nitems)){
					wrd[j].count=bit_count(wrd[j].count);
					if(w_id)
						j++;
					else
						continue;
				}else{
					/* Skip this result */
				}
			}else{
				if(w_id)
					j++;
				else
					continue;
			}
			wrd[j]=wrd[i];
		}
	}
	/* Check last word */
	switch(query->search_mode){
		case UDM_MODE_ALL:
			query->total_found=(wrd[j].count>=words_in_query)?j+1:j;
			break;
		case UDM_MODE_BOOL:
			for(item=0;item<nitems;item++){
				if(temp_items[item].cmd==UDM_STACK_WORD){
					/* Change word mask by 1 or 0 */
					temp_items[item].arg=((items[item].arg)&(wrd[j].count))?1:0;
				}
			}
			wrd[j].count=bit_count(wrd[j].count);
			query->total_found=(UdmCalcBoolItems(temp_items,nitems))?j+1:j;
			break;
		case UDM_MODE_ANY:
		default:
			query->total_found=j+1;
			break;
	}
	return(wrd);
}
